"""
Defines how OpenMDAO variables are serialized to XML using a conversion table
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import logging
import re
from typing import IO, Union

import numpy as np
from lxml import etree
from lxml.etree import (
    XPathEvalError,
    _Comment,
    _Element,
)  # pylint: disable=protected-access  # Useful for type hinting
from openmdao.vectors.vector import Vector

from fastoad._utils.files import make_parent_dir
from fastoad._utils.strings import get_float_list_from_string
from fastoad.io.formatter import IVariableIOFormatter
from fastoad.io.xml.exceptions import (
    FastXPathEvalError,
    FastXmlFormatterDuplicateVariableError,
    FastXpathTranslatorVariableError,
    FastXpathTranslatorXPathError,
)
from fastoad.io.xml.translator import VarXpathTranslator
from fastoad.openmdao.variables import VariableList
from .constants import DEFAULT_IO_ATTRIBUTE, DEFAULT_UNIT_ATTRIBUTE, ROOT_TAG

_LOGGER = logging.getLogger(__name__)  # Logger for this module


class VariableXmlBaseFormatter(IVariableIOFormatter):
    """
    Customizable formatter for variables

    User must provide at instantiation a VarXpathTranslator instance that tells how
    variable names should be converted from/to XPath.

    Note: XPath are always considered relatively to the root. Therefore, "foo/bar" should be
    provided to match following XML structure:

    .. code-block:: xml

        <root>
            <foo>
                <bar>
                    "some value"
                </bar>
            </foo>
        </root>

    :param translator: the VarXpathTranslator instance
    """

    def __init__(self, translator: VarXpathTranslator):
        self._translator = translator

        #: The XML attribute key for specifying units
        self.xml_unit_attribute = DEFAULT_UNIT_ATTRIBUTE

        #: The XML attribute key for specifying I/O status
        self.xml_io_attribute = DEFAULT_IO_ATTRIBUTE

        #: Used for converting read units in units recognized by OpenMDAO.
        #  Regular expressions can be used in dict keys.
        self.unit_translation = {
            "²": "**2",
            "³": "**3",
            "°": "deg",
            "°C": "degC",
            "kt": "kn",
            "\bin\b": "inch",
        }

    def read_variables(self, data_source: Union[str, IO]) -> VariableList:
        variables = VariableList()

        # If there is a comment, it will be used as description if the previous
        # element described a variable.
        previous_variable_name = None

        parser = etree.XMLParser(remove_blank_text=True, remove_comments=False)
        tree = etree.parse(data_source, parser)
        root = tree.getroot()
        for elem in root.iter():
            if isinstance(elem, _Comment) and previous_variable_name is not None:
                variables[previous_variable_name].description = elem.text.strip()
                previous_variable_name = None
                continue

            units = elem.attrib.get(self.xml_unit_attribute, None)
            is_input = elem.attrib.get(self.xml_io_attribute, None)
            if units:
                # Ensures compatibility with OpenMDAO units
                for legacy_chars, om_chars in self.unit_translation.items():
                    units = re.sub(legacy_chars, om_chars, units)
                    units = units.replace(legacy_chars, om_chars)
            value = None
            if elem.text:
                value = get_float_list_from_string(elem.text)

            previous_variable_name = None
            if value is not None:
                try:
                    path_tags = [ancestor.tag for ancestor in elem.iterancestors()]
                    path_tags.reverse()
                    path_tags.append(elem.tag)
                    xpath = "/".join(path_tags[1:])  # Do not use root tag
                    name = self._translator.get_variable_name(xpath)
                except FastXpathTranslatorXPathError as err:
                    _LOGGER.warning(
                        "The xpath %s does not have any variable " "affected in the translator.",
                        err.xpath,
                    )
                    continue

                if name not in variables.names():
                    # Add Variable
                    if is_input is not None:
                        is_input = is_input == "True"

                    variables[name] = {"val": value, "units": units, "is_input": is_input}
                    previous_variable_name = name
                else:
                    raise FastXmlFormatterDuplicateVariableError(
                        "Variable %s is defined in more than one place in file %s"
                        % (name, data_source)
                    )

        return variables

    def write_variables(self, data_source: Union[str, IO], variables: VariableList):

        root = etree.Element(ROOT_TAG)

        for variable in variables:

            try:
                xpath = self._translator.get_xpath(variable.name)
            except FastXpathTranslatorVariableError as exc:
                _LOGGER.warning("No translation found: %s", exc)
                continue
            element = self._create_xpath(root, xpath)

            # Set value, units and io
            if variable.units:
                element.attrib[self.xml_unit_attribute] = variable.units
            if variable.is_input is not None:
                element.attrib[self.xml_io_attribute] = str(variable.is_input)

            # Filling value for already created element
            element.text = str(variable.value)
            if not isinstance(variable.value, (np.ndarray, Vector, list)):
                # Here, it should be a float
                element.text = str(variable.value)
            elif len(np.squeeze(variable.value).shape) == 0:
                element.text = str(np.squeeze(variable.value).item())
            else:
                element.text = json.dumps(np.asarray(variable.value).tolist())
            if variable.description:
                element.append(etree.Comment(variable.description))
        # Write
        tree = etree.ElementTree(root)
        make_parent_dir(data_source)
        tree.write(data_source, pretty_print=True)

    @staticmethod
    def _create_xpath(root: _Element, xpath: str) -> _Element:
        """
        Creates required XML Path from provided root element

        :param root:
        :param xpath:
        :return: created element
        """
        if xpath.startswith("/"):
            xpath = xpath[1:]  # needed to avoid empty string at first place after split
        path_components = xpath.split("/")
        element = root
        children = []
        # Create XML path if needed
        for path_component in path_components:
            try:
                children = element.xpath(path_component)
            except XPathEvalError:
                raise FastXPathEvalError('Could not resolve XPath "%s"' % path_component)
            if not children:
                # Build path
                new_element = etree.Element(path_component)
                element.append(new_element)
                element = new_element
            else:
                # Use existing path
                # (Unicity of OpenMDAO variables makes that children should not have more that
                # one element)
                element = children[0]

        return element



from .geometry.geometry import ComputeGeometry
from .aerodynamics.aerodynamics import ComputeAerodynamics
from .mass.mass import ComputeMass
from .performance.performance import ComputePerformance
from .update_mtow.update_mtow import UpdateMTOW

import openmdao.api as om

import fastoad.api as oad


@oad.RegisterOpenMDAOSystem("tutorial.fast_oad.global")
class SizingLoopMTOW(om.Group):
    """
    Gather all the discipline module/groups into the main problem
    """

    def setup(self):

        self.add_subsystem(name="compute_geometry", subsys=ComputeGeometry(), promotes=["*"])
        self.add_subsystem(
            name="compute_aerodynamics", subsys=ComputeAerodynamics(), promotes=["*"]
        )
        self.add_subsystem(name="compute_mass", subsys=ComputeMass(), promotes=["*"])
        self.add_subsystem(name="compute_performance", subsys=ComputePerformance(), promotes=["*"])
        self.add_subsystem(name="update_mtow", subsys=UpdateMTOW(), promotes=["*"])



"""
Computation of beam linear weight.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om

import fastoad.api as oad


@oad.RegisterOpenMDAOSystem("tutorial.beam_problem.weight")
class LinearWeight(om.ExplicitComponent):
    """
    Computes beam linear weight given its dimensions.
    """

    def setup(self):

        self.add_input("data:geometry:l", val=np.nan, units="m")
        self.add_input("data:geometry:h", val=np.nan, units="m")
        self.add_input("data:material:density", val=np.nan, units="kg/m**3")

        self.add_output("data:weight:linear_weight", val=10.0, units="N/m")

    def setup_partials(self):

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):

        l = inputs["data:geometry:l"]
        h = inputs["data:geometry:h"]
        rho = inputs["data:material:density"]

        outputs["data:weight:linear_weight"] = l * h * rho * 9.81



import numpy as np


def compute_k(aspect_ratio):
    """
    Computes the induced drag coefficient based on the aspect ratio

    :param aspect_ratio: Wing aspect ratio, no unit

    :return k: induced drag coefficient, no unit
    """

    # Computation of the Oswald efficiency factor
    e = 1.78 * (1.0 - 0.045 * aspect_ratio ** 0.68) - 0.64

    # Computation of the lift induced drag coefficient
    k = 1.0 / (np.pi * aspect_ratio * e)

    return k



import numpy as np

import openmdao.api as om


class ComputeOweExercise(om.ExplicitComponent):
    """
    Computes the aircraft structural mass based on its MTOW and wing mass
    """

    def setup(self):

        # Defining the input(s)

        # Defining the output(s)

        pass

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        pass



from .sub_components.compute_profile_drag import compute_cd0
from .sub_components.compute_induced_drag_coefficient import compute_k
from .sub_components.compute_lift_to_drag_ratio import compute_l_d


def compute_aerodynamics(cruise_altitude, cruise_speed, mtow, wing_area, aspect_ratio):
    """
    Gather all the aerodynamics sub-functions in the main function

    :param cruise_altitude: Cruise altitude, in m
    :param cruise_speed: Cruise speed, in m/s
    :param mtow: Max Take-Off Weight, in kg
    :param wing_area: Wing area; in m2
    :param aspect_ratio: Wing aspect ratio, no unit

    :return wing_area: Wing area, in m2
    """

    # Let's start by computing the profile drag coefficient
    cd0 = compute_cd0(wing_area=wing_area)

    # Let's now compute the lift induced drag coefficient
    k = compute_k(aspect_ratio=aspect_ratio)

    # We can now compute the lift-to-drag ratio
    l_d = compute_l_d(
        cruise_altitude=cruise_altitude,
        cruise_speed=cruise_speed,
        cd0=cd0,
        k=k,
        mtow=mtow,
        wing_area=wing_area,
    )

    return l_d



"""
Classes for computation of routes (i.e. assemblies of climb, cruise and descent phases).
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from scipy.optimize import root_scalar

from fastoad.model_base import FlightPoint
from fastoad.model_base.datacls import MANDATORY_FIELD
from fastoad.models.performances.mission.base import FlightSequence, IFlightPart
from fastoad.models.performances.mission.segments.base import AbstractFlightSegment
from fastoad.models.performances.mission.segments.cruise import CruiseSegment


@dataclass
class SimpleRoute(FlightSequence):
    """
    Computes a simple route.

    The computed route will be made of:
        - any number of climb phases
        - one cruise segment
        - any number of descent phases.
    """

    #: Any number of flight phases that will occur before cruise.
    climb_phases: List[FlightSequence] = MANDATORY_FIELD

    #: The cruise phase.
    cruise_segment: CruiseSegment = MANDATORY_FIELD

    #: Any number of flight phases that will occur after cruise.
    descent_phases: List[FlightSequence] = MANDATORY_FIELD

    def __post_init__(self):
        super().__post_init__()
        self.flight_sequence.extend(self._get_flight_sequence())

    @property
    def cruise_distance(self):
        """
        Ground distance to be covered during cruise, as set in target of :attr:`cruise_segment`.
        """
        return self.cruise_segment.target

    @cruise_distance.setter
    def cruise_distance(self, cruise_distance):
        self.cruise_segment.target.ground_distance = float(cruise_distance)
        self.cruise_segment.target.set_as_relative("ground_distance")

    @property
    def cruise_speed(self) -> Optional[Tuple[str, float]]:
        """
        Type (among `true_airspeed`, `equivalent_airspeed` and `mach`) and value of cruise speed.
        """
        climb_segments = []
        for phase in self.climb_phases:
            climb_segments += phase.flight_sequence

        climb_segments.reverse()
        for segment in climb_segments:
            for speed_param in ["true_airspeed", "equivalent_airspeed", "mach"]:
                speed_value = getattr(segment.target, speed_param)
                if speed_value and speed_value != AbstractFlightSegment.CONSTANT_VALUE:
                    return speed_param, speed_value

        return None

    def _get_flight_sequence(self) -> List[IFlightPart]:
        # The preliminary climb segment of the cruise segment is set to the
        # last segment before cruise.
        cruise_climb = self.climb_phases[-1]
        while isinstance(cruise_climb, FlightSequence):
            cruise_climb = cruise_climb.flight_sequence[-1]
        self.cruise_segment.climb_segment = cruise_climb

        return self.climb_phases + [self.cruise_segment] + self.descent_phases


@dataclass
class RangedRoute(SimpleRoute):
    """
    Computes a route so that it covers the specified ground distance.
    """

    #: Target ground distance for whole route
    flight_distance: float = MANDATORY_FIELD

    #: Accuracy on actual total ground distance for the solver. In meters
    distance_accuracy: float = 0.5e3

    def __post_init__(self):
        super().__post_init__()

        # We will use this to keep data along root_scalar process (see _solve_cruise_distance() )
        self._flight_points = None

    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        # In very simple cases, climb and descent phases can have fixed
        # covered ground distance. In that case, cruise distance is easy to
        # obtain from flight_distance.
        # In other cases, cruise distance is obtained using a solver.
        climb_descent_distances = []
        for phase in self.climb_phases + self.descent_phases:
            climb_descent_distances.extend(self._get_ground_distances(phase))

        if 0.0 in climb_descent_distances:
            return self._solve_cruise_distance(start)

        self.cruise_distance = self.flight_distance - np.sum(climb_descent_distances)
        return super().compute_from(start)

    @classmethod
    def _get_ground_distances(cls, phase: FlightSequence) -> list:
        ground_distances = []
        for flight_part in phase.flight_sequence:
            if isinstance(flight_part, AbstractFlightSegment):
                ground_distances.append(flight_part.target.ground_distance)
            else:
                ground_distances.extend(cls._get_ground_distances(flight_part))

        return ground_distances

    def _solve_cruise_distance(self, start: FlightPoint) -> pd.DataFrame:
        """
        Adjusts cruise distance through a solver to have whole route that
        matches provided flight distance.
        """

        root_scalar(
            self._compute_flight,
            args=(start,),
            x0=self.flight_distance * 0.5,
            x1=self.flight_distance * 0.25,
            xtol=self.distance_accuracy,
            method="secant",
        )

        return self._flight_points

    def _compute_flight(self, cruise_distance, start: FlightPoint):
        """
        Computes flight for provided cruise distance

        :param cruise_distance:
        :param start:
        :return: difference between computes distance and self.flight_distance
        """
        self.cruise_distance = cruise_distance
        self._flight_points = super().compute_from(start)
        obtained_distance = (
            self._flight_points.iloc[-1].ground_distance
            - self._flight_points.iloc[0].ground_distance
        )
        return self.flight_distance - obtained_distance



"""Checks about packages and resources"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from importlib.resources import contents
from typing import List


class PackageReader:
    """
    Wrapper of `importlib.resources.contents` which handles when `package_name` is
    not a package or does not exist.

    :param package_name: Name of package to inspect.
    """

    def __init__(self, package_name: str):
        """
        :param package_name:
        """
        self.is_package = False
        self.exists = True
        self.is_module = False
        self.has_error = False
        self._contents = []
        self.package_name = package_name

    @property
    def package_name(self):
        """Name of package to inspect."""
        return self._package_name

    @package_name.setter
    def package_name(self, package_name):
        self._package_name = package_name
        if package_name:
            try:
                self._contents = list(contents(package_name))
                self.is_package = True
            except TypeError:
                self.is_module = True
            except ModuleNotFoundError:
                # Either the indicated package does not exist, or it is a file with no extension.
                # We want to ensure existence to correctly set self.exists.
                if "." in package_name:
                    parent_package_name, module_name = package_name.rsplit(".", 1)
                    parent_package = PackageReader(parent_package_name)
                    if parent_package.is_package and module_name in parent_package.contents:
                        self.exists = True
                else:
                    # Here we assume non-existence
                    self.exists = False
            except Exception:  # pylint: disable = W0703
                # Here we catch any Python error that may happen when reading the loaded code.
                # Thus, we ensure to not break the application if a module is incorrectly written.
                self.has_error = True

    @property
    def contents(self) -> List[str]:
        """
        An iterator if package exists, otherwise an empty list.
        """
        return self._contents



import openmdao.api as om

from .sub_components.compute_fuel_mass import ComputeFuelMass


class ComputePerformance(om.Group):
    def setup(self):

        self.add_subsystem(name="compute_fuel_mass", subsys=ComputeFuelMass(), promotes=["*"])



import numpy as np
import scipy.constants as sc

import openmdao.api as om


class ComputeWingMass(om.ExplicitComponent):
    """
    Computes the wing mass based on the MTOW, its area and aspect ratio
    """

    def setup(self):

        # Defining the input(s)

        self.add_input(name="wing_area", units="ft**2", val=np.nan)
        # Notice that here we ask for the wing area in sq. ft as it is the unit we need for the
        # formula, so we won't need to convert the wing area in the proper unit
        self.add_input(name="mtow", units="lbm", val=np.nan)
        # Same for the MTOW
        self.add_input(name="aspect_ratio", val=np.nan)

        # Defining the output(s)

        self.add_output(name="wing_mass", units="lbm")
        # Same situation here, the formula outputs in lbm but if we later want to use it in kg,
        # we will just have to ask for units="kg" and OpenMDAO automatically handles the conversion

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        # Assigning the input to local variable for clarity
        wing_area = inputs["wing_area"]
        aspect_ratio = inputs["aspect_ratio"]
        mtow = inputs["mtow"]

        # Let's now apply the formula
        wing_mass = (
            96.948
            * (
                (5.7 * mtow / 1.0e5) ** 0.65
                * aspect_ratio ** 0.57
                * (wing_area / 100.0) ** 0.61
                * 2.5
            )
            ** 0.993
        )

        outputs["wing_mass"] = wing_mass



import openmdao.api as om

from .sub_components.compute_wing_mass import ComputeWingMass
from .sub_components.compute_owe_exercise import ComputeOweExercise
from .sub_components.compute_owe import ComputeOwe


class ComputeMassCorrect(om.Group):
    def setup(self):

        self.add_subsystem(name="compute_wing_mass", subsys=ComputeWingMass(), promotes=["*"])
        self.add_subsystem(name="compute_owe", subsys=ComputeOwe(), promotes=["*"])


class ComputeMassExercise(om.Group):
    def setup(self):

        self.add_subsystem(name="compute_wing_mass", subsys=ComputeWingMass(), promotes=["*"])
        self.add_subsystem(name="compute_owe", subsys=ComputeOweExercise(), promotes=["*"])



"""
Defines how OpenMDAO variables are serialized to XML
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
from typing import Union, IO

from fastoad.openmdao.variables import VariableList
from .exceptions import FastXPathEvalError
from .translator import VarXpathTranslator
from .variable_io_base import VariableXmlBaseFormatter

_LOGGER = logging.getLogger(__name__)  # Logger for this module


class VariableXmlStandardFormatter(VariableXmlBaseFormatter):
    """
    Standard XML formatter for variables

    Assuming self.path_separator is defined as ``:`` (default), a variable named like
    ``foo:bar`` with units ``m/s`` will be read and written as:

    .. code:: xml

        <aircraft>
            <foo>
                <bar units="m/s" >`42.0</bar>
            </foo>
        <aircraft>

    When writing outputs of a model, OpenMDAO component hierarchy may be used by defining

     .. code:: python

        self.path_separator = '.'  # Discouraged for reading !
        self.use_promoted_names = False

    This way, a variable like ``componentA.subcomponent2.my_var`` will be written as:

    .. code:: xml

        <aircraft>
            <componentA>
                <subcomponent2>
                    <my_var units="m/s" >72.0</my_var>
                </subcomponent2>
            <componentA>
        <aircraft>
    """

    def __init__(self):
        super().__init__(BasicVarXpathTranslator(":"))

    @property
    def path_separator(self):
        """
        The separator that will be used in OpenMDAO variable names to match XML path.
        Warning: The dot "." can be used when writing, but not when reading.
        """
        return self._translator.path_separator

    @path_separator.setter
    def path_separator(self, separator):
        self._translator.path_separator = separator

    def read_variables(self, data_source: Union[str, IO]) -> VariableList:
        # Check separator, as OpenMDAO won't accept the dot.
        if self.path_separator == ".":
            _LOGGER.warning(
                "Usage of dot (.) is discouraged, because it won't be accepted by OpenMDAO."
            )
        return super().read_variables(data_source)

    def write_variables(self, data_source: Union[str, IO], variables: VariableList):
        try:
            super().write_variables(data_source, variables)
        except FastXPathEvalError as err:
            # Trying to help...
            raise FastXPathEvalError(
                err.args[0] + ' : self.path_separator is "%s". It is correct?' % self.path_separator
            )


class BasicVarXpathTranslator(VarXpathTranslator):
    """
    Dedicated VarXpathTranslator that builds variable names by simply converting
    the '/' separator of XPaths into the desired separator.
    """

    def __init__(self, path_separator):
        super().__init__()
        self.path_separator = path_separator

    def get_variable_name(self, xpath: str) -> str:
        path_components = xpath.split("/")
        name = self.path_separator.join(path_components)
        return name

    def get_xpath(self, var_name: str) -> str:
        path_components = var_name.split(self.path_separator)
        xpath = "/".join(path_components)
        return xpath



import numpy as np


def compute_owe(wing_mass, mtow):
    """
    Computes the aircraft structural mass based on its MTOW and wing mass

    :param wing_mass: Wing mass, in kg
    :param mtow: Max Take-Off Weight, in kg

    :return owe: the structural mass, in kg
    """

    # Let's start by computing the weight of the aircraft without the wings
    owe_without_wing = mtow * (0.43 + 0.0066 * np.log(mtow))

    # Let's now add the wing mass to get the structural weight
    owe = owe_without_wing + wing_mass

    return owe



"""Structure for managing flight point data."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from copy import deepcopy
from dataclasses import asdict, dataclass, fields
from numbers import Number
from typing import Any, List, Mapping, Sequence, Union

import numpy as np
import pandas as pd

from fastoad.constants import EngineSetting


@dataclass
class FlightPoint:
    """
    Dataclass for storing data for one flight point.

    This class is meant for:

    - pandas friendliness: data exchange with pandas DataFrames is simple
    - extensibility: any user might add fields to the **class** using :meth:`add_field`

    **Exchanges with pandas DataFrame**

        A pandas DataFrame can be generated from a list of FlightPoint instances::

            >>> import pandas as pd
            >>> from fastoad.model_base import FlightPoint

            >>> fp1 = FlightPoint(mass=70000., altitude=0.)
            >>> fp2 = FlightPoint(mass=60000., altitude=10000.)
            >>> df = pd.DataFrame([fp1, fp2])

        And FlightPoint instances can be created from DataFrame rows::

            # Get one FlightPoint instance from a DataFrame row
            >>> fp1bis = FlightPoint.create(df.iloc[0])

            # Get a list of FlightPoint instances from the whole DataFrame
            >>> flight_points = FlightPoint.create_list(df)

    **Extensibility**

        FlightPoint class is bundled with several fields that are commonly used in trajectory
        assessment, but one might need additional fields.

        Python allows to add attributes to any instance at runtime, but for FlightPoint to run
        smoothly, especially when exchanging data with pandas, you have to work at class level.
        This can be done using :meth:`add_field`, preferably outside of any class or function::

            # Adds a float field with None as default value
            >>> FlightPoint.add_field("ion_drive_power")

            # Adds a field and define its type and default value
            >>> FlightPoint.add_field("warp", annotation_type=int, default_value=9)

            # Now these fields can be used at instantiation
            >>> fp = FlightPoint(ion_drive_power=110.0, warp=12)

            # Removes a field, even an original one (useful only to avoid having it in outputs)
            >>> FlightPoint.remove_field("sfc")

    .. note::

        All parameters in FlightPoint instances are expected to be in SI units.

    """

    time: float = 0.0  #: Time in seconds.
    altitude: float = None  #: Altitude in meters.
    ground_distance: float = 0.0  #: Covered ground distance in meters.
    mass: float = None  #: Mass in kg.
    true_airspeed: float = None  #: True airspeed (TAS) in m/s.
    equivalent_airspeed: float = None  #: Equivalent airspeed (EAS) in m/s.
    mach: float = None  #: Mach number.
    engine_setting: EngineSetting = None  #: Engine setting.
    #: Lift coefficient.
    CL: float = None  # pylint: disable=invalid-name
    #: Drag coefficient.
    CD: float = None  # pylint: disable=invalid-name
    drag: float = None  #: Aircraft drag in Newtons.
    thrust: float = None  #: Thrust in Newtons.
    thrust_rate: float = None  #: Thrust rate (between 0. and 1.)

    #: If True, propulsion should match the thrust value.
    #: If False, propulsion should match thrust rate.
    thrust_is_regulated: bool = None

    sfc: float = None  #: Specific Fuel Consumption in kg/N/s.
    slope_angle: float = None  #: Slope angle in radians.
    acceleration: float = None  #: Acceleration value in m/s**2.
    name: str = None  #: Name of current phase.

    _units = dict(
        time="s",
        altitude="m",
        ground_distance="m",
        mass="kg",
        true_airspeed="m/s",
        equivalent_airspeed="m/s",
        mach="-",
        CL="-",
        CD="-",
        drag="N",
        thrust="N",
        thrust_rate="-",
        sfc="kg/N/s",
        slope_angle="rad",
        acceleration="m/s**2",
    )

    def __post_init__(self):
        self._relative_parameters = {"ground_distance", "time"}

    def set_as_relative(self, field_names: Union[Sequence[str], str]):
        """
        Makes that values for given field_names will be considered as relative when
        calling :meth:`make_absolute`.

        :param field_names:
        """
        if isinstance(field_names, str):
            self._relative_parameters.add(field_names)
        else:
            self._relative_parameters |= set(field_names)

    def set_as_absolute(self, field_names: Union[Sequence[str], str]):
        """
        Makes that values for given field_names will be considered as absolute when
        calling :meth:`make_absolute`.

        :param field_names:
        """
        if isinstance(field_names, str):
            self._relative_parameters.remove(field_names)
        else:
            self._relative_parameters -= set(field_names)

    def is_relative(self, field_name) -> bool:
        """
        Tells if given field is considered as relative or absolut

        :param field_name:
        :return: True if it is relative
        """
        return field_name in self._relative_parameters

    def make_absolute(self, reference_point: "FlightPoint") -> "FlightPoint":
        """
        Computes a copy flight point where no field is relative.

        :param reference_point: relative fields will be made absolute using this point.
        :return: the copied flight point with no relative field.
        """
        new_point = deepcopy(self)
        for field in fields(new_point):
            reference_value = getattr(reference_point, field.name)
            target_value = getattr(new_point, field.name)
            if isinstance(target_value, Number) and new_point.is_relative(field.name):
                setattr(new_point, field.name, reference_value + target_value)
                new_point.set_as_absolute(field.name)
        new_point.scalarize()
        return new_point

    @classmethod
    def get_field_names(cls):
        """
        :return: names of all fields of the flight point.
        """
        return [field.name for field in fields(cls) if not field.name.startswith("_")]

    @classmethod
    def get_units(cls) -> dict:
        """
        Returns (field name, unit) dict for any field that has a defined unit.

        A dimensionless physical quantity will have "-" as unit.
        """
        return cls._units

    @classmethod
    def create(cls, data: Mapping) -> "FlightPoint":
        """
        Instantiate FlightPoint from provided data.

        `data` can typically be a dict or a pandas DataFrame row.

        :param data: a dict-like instance where keys are FlightPoint attribute names
        :return: the created FlightPoint instance
        """
        return cls(**dict(data))

    @classmethod
    def create_list(cls, data: pd.DataFrame) -> List["FlightPoint"]:
        """
        Creates a list of FlightPoint instances from provided DataFrame.

        :param data: a dict-like instance where keys are FlightPoint attribute names
        :return: the created FlightPoint instance
        """
        return [cls.create(row) for row in data.iloc]

    @classmethod
    def add_field(cls, name: str, annotation_type=float, default_value: Any = None, unit=None):
        """
        Adds the named field to FlightPoint class.

        If the field name already exists, the field is redefined.

        :param name: field name
        :param annotation_type: field type
        :param default_value: field default value
        :param unit: expected unit for the added field ("-" should be provided for a dimensionless
                     physical quantity)
        """
        cls.remove_field(name)

        del cls.__init__  # Delete constructor to allow it being rebuilt with dataclass() call
        setattr(cls, name, default_value)
        cls.__annotations__[name] = annotation_type
        dataclass(cls)
        if unit:
            cls._units[name] = unit

    @classmethod
    def remove_field(cls, name):
        """
        Removes the named field from FlightPoint class.

        :param name: field name
        """
        if name in cls.__annotations__:
            del cls.__init__  # Delete constructor to allow it being rebuilt with dataclass() call
            delattr(cls, name)
            del cls.__annotations__[name]
            dataclass(cls)
            if name in cls._units:
                del cls._units[name]

    def scalarize(self):
        """
        Convenience method for converting to scalars all fields that have a
        one-item array-like value.
        """
        self_as_dict = asdict(self)
        for field_name, value in self_as_dict.items():
            if np.size(value) == 1:
                setattr(self, field_name, np.asarray(value).item())



"""
FAST-OAD model for mission computation.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from importlib.resources import path

import openmdao.api as om
import pandas as pd

from fastoad.module_management.constants import ModelDomain
from fastoad.module_management.service_registry import RegisterOpenMDAOSystem
from . import resources
from .mission_component import MissionComponent, _get_variable_name_provider
from .mission_wrapper import MissionWrapper
from ..mission_definition.schema import MissionDefinition

_LOGGER = logging.getLogger(__name__)  # Logger for this module


@RegisterOpenMDAOSystem("fastoad.performances.mission", domain=ModelDomain.PERFORMANCE)
class Mission(om.Group):
    """
    Computes a mission as specified in mission input file.
    """

    def initialize(self):
        self.options.declare(
            "propulsion_id",
            default="",
            types=str,
            desc="(mandatory) The identifier of the propulsion wrapper.",
        )
        self.options.declare(
            "out_file",
            default="",
            types=str,
            desc="If provided, a csv file will be written at provided path with all computed "
            "flight points.",
        )
        self.options.declare(
            "mission_file_path",
            default="::sizing_mission",
            types=(str, MissionDefinition),
            allow_none=True,
            desc="The path to file that defines the mission.\n"
            'If can also begin with two colons "::" to use pre-defined missions:\n'
            '  - "::sizing_mission" : design mission for CeRAS-01\n'
            '  - "::breguet" : a simple mission with Breguet formula for cruise, and input\n'
            "    coefficients for fuel reserve and fuel consumption during climb and descent",
        )
        self.options.declare(
            "mission_name",
            default=None,
            types=str,
            allow_none=True,
            desc="The mission name. Required if mission file defines several missions.",
        )
        self.options.declare(
            "use_initializer_iteration",
            default=True,
            types=bool,
            desc="During first solver loop, a complete mission computation can fail or consume "
            "useless CPU-time.\n"
            "When activated, this option ensures the first iteration is done using a simple,\n"
            "dummy, formula instead of the specified mission.\n"
            "Set this option to False if you do expect this model to be computed only once.",
        )
        self.options.declare(
            "adjust_fuel",
            default=True,
            types=bool,
            desc="If True, block fuel will fit fuel consumption during mission. In that case, a "
            "solver (local or global) will be needed. (see `add_solver` option for more"
            "information)\n"
            "If False, block fuel will be taken from input data.",
        )
        self.options.declare(
            "compute_input_weight",
            default=False,
            types=bool,
            desc="If True, input weight will be deduced from block fuel.\n"
            "If False, block fuel will be deduced from input weight.\n"
            "Not used (actually forced to True) if adjust_fuel is True.",
        )
        self.options.declare(
            "compute_TOW",
            default=False,
            types=bool,
            desc="(Deprecated. Replaced by compute_input_weight)\n"
            "If True, TakeOff Weight will be computed from onboard fuel at takeoff and ZFW.\n"
            "If False, block fuel will be computed from ramp weight and ZFW.\n"
            "Not used (actually forced to True) if adjust_fuel is True.",
        )
        self.options.declare(
            "add_solver",
            default=False,
            types=bool,
            desc="If True, a local solver is set for the mission computation.\n"
            "It is useful if `adjust_fuel` is set to True, or to ensure consistency between "
            "ramp weight and takeoff weight + taxi-out fuel.\n"
            "If a global solver is used, using a local solver or not should be only a question"
            "of CPU time consumption and is not expected to modify the results.",
        )
        self.options.declare(
            "is_sizing",
            default=False,
            types=bool,
            desc="If True, TOW will be considered equal to MTOW and mission payload will be "
            "considered equal to design payload.",
        )
        self.options.declare(
            "reference_area_variable",
            default="data:geometry:wing:area",
            types=str,
            desc="Defines the name of the variable for providing aircraft reference surface area.",
        )

    def setup(self):
        self.options["compute_input_weight"] = self.options["compute_TOW"]

        if "::" in self.options["mission_file_path"]:
            # The configuration file parser will have added the working directory before
            # the file name. But as the user-provided string begins with "::", we just
            # have to ignore all before "::".
            i = self.options["mission_file_path"].index("::")
            file_name = self.options["mission_file_path"][i + 2 :] + ".yml"
            with path(resources, file_name) as mission_input_file:
                self.options["mission_file_path"] = MissionDefinition(mission_input_file)
        self._mission_wrapper = MissionWrapper(self.options["mission_file_path"])
        if self.options["mission_name"] is None:
            self.options["mission_name"] = self._mission_wrapper.get_unique_mission_name()

        mission_name = self.options["mission_name"]
        self._name_provider = _get_variable_name_provider(mission_name)

        self.add_subsystem("ZFW_computation", self._get_zfw_component(mission_name), promotes=["*"])

        if self.options["adjust_fuel"]:
            self.options["compute_input_weight"] = True
            self.connect(
                self._name_provider.NEEDED_BLOCK_FUEL.value,
                self._name_provider.BLOCK_FUEL.value,
            )
        if self.options["add_solver"]:
            self.nonlinear_solver = om.NonlinearBlockGS(maxiter=30, rtol=1.0e-4, iprint=0)
            self.linear_solver = om.DirectSolver()

        if self.options["compute_input_weight"]:
            self.add_subsystem(
                "input_mass_computation",
                self._get_input_weight_component(mission_name),
                promotes=["*"],
            )

        mission_options = dict(self.options.items())
        del mission_options["adjust_fuel"]
        del mission_options["compute_input_weight"]
        del mission_options["compute_TOW"]
        del mission_options["add_solver"]
        del mission_options["mission_file_path"]
        mission_options["mission_wrapper"] = self._mission_wrapper
        mission_options["mission_name"] = mission_name

        self.add_subsystem(
            "mission_computation", MissionComponent(**mission_options), promotes=["*"]
        )
        if not self.options["compute_input_weight"]:
            self.add_subsystem(
                "block_fuel_computation",
                self._get_block_fuel_component(mission_name),
                promotes=["*"],
            )

    @property
    def flight_points(self) -> pd.DataFrame:
        """Dataframe that lists all computed flight point data."""
        return self.mission_computation.flight_points

    def _get_zfw_component(self, mission_name: str) -> om.AddSubtractComp:
        """

        :param mission_name:
        :return: component that computes Zero Fuel Weight from OWE and mission payload
        """
        if self.options["is_sizing"]:
            payload_var = "data:weight:aircraft:payload"
        else:
            payload_var = self._name_provider.PAYLOAD.value

        zfw_computation = om.AddSubtractComp()
        zfw_computation.add_equation(
            self._name_provider.ZFW.value,
            ["data:weight:aircraft:OWE", payload_var],
            units="kg",
            desc=f'Zero Fuel Weight for mission "{mission_name}"',
        )
        return zfw_computation

    def _get_input_weight_component(self, mission_name: str):
        input_weight_variable = self._mission_wrapper.get_input_weight_variable_name(mission_name)
        if not input_weight_variable:
            return None

        computation = om.AddSubtractComp()
        computation.add_equation(
            output_name=input_weight_variable,
            input_names=[
                self._name_provider.ZFW.value,
                self._name_provider.BLOCK_FUEL.value,
                self._name_provider.CONSUMED_FUEL_BEFORE_INPUT_WEIGHT.value,
            ],
            units="kg",
            scaling_factors=[1, 1, -1],
            desc=f'Loaded fuel at beginning for mission "{mission_name}"',
        )

        return computation

    def _get_block_fuel_component(self, mission_name: str) -> om.AddSubtractComp:
        """

        :param mission_name:
        :return: component that computes block fuel from ramp weight and ZFW
        """
        input_weight_variable = self._mission_wrapper.get_input_weight_variable_name(mission_name)

        block_fuel_computation = om.AddSubtractComp()
        block_fuel_computation.add_equation(
            output_name=self._name_provider.BLOCK_FUEL.value,
            input_names=[
                input_weight_variable,
                self._name_provider.CONSUMED_FUEL_BEFORE_INPUT_WEIGHT.value,
                self._name_provider.ZFW.value,
            ],
            units="kg",
            scaling_factors=[1, 1, -1],
            desc=f'Loaded fuel at beginning for mission "{mission_name}"',
        )

        return block_fuel_computation



"""
Helper module for copying resources
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
import shutil
from importlib.resources import Package, is_resource, path
from types import ModuleType
from typing import List

from .contents import PackageReader
from ..files import make_parent_dir


def copy_resource(package: Package, resource: str, target_path):
    """
    Copies the indicated resource file to provided target path.

    If target_path matches an already existing folder, resource file will be copied in this folder
    with same name. Otherwise, target_path will be the used name of copied resource file

    :param package: package as in importlib.resources.read_binary()
    :param resource: resource as in importlib.resources.read_binary()
    :param target_path: file system path
    """
    make_parent_dir(target_path)

    with path(package, resource) as source_path:
        shutil.copy(source_path, target_path)


def copy_resource_folder(package: Package, destination_path: str, exclude: List[str] = None):
    """
    Copies the full content of provided package in destination folder.

    Names of files that should not be copied have to be listed in `exclude`.

    .. Warning ::

        As the resources in the folder are discovered by browsing
        through the folder, they are not explicitly listed in the Python code.
        Therefore, to have the install process run smoothly, these resources need
        to be listed in the MANIFEST.in file.

    :param package: name of resource package to copy
    :param destination_path: file system path of destination
    :param exclude: list of item names that should not be copied
    """
    exclusion_list = ["__pycache__"]
    if exclude:
        exclusion_list += exclude

    package_contents = PackageReader(package).contents
    if package_contents:
        for resource_name in package_contents:
            if resource_name in exclusion_list:
                continue
            if is_resource(package, resource_name):
                destination_file_path = pth.join(destination_path, resource_name)
                copy_resource(package, resource_name, destination_file_path)
            else:
                # In case of subfolders that are only declared in MANIFEST.in,
                # getattr(package, "resource_name") will fail (is there another way?).
                # So we fall back to using package name as as string.
                if isinstance(package, ModuleType):
                    package_name = package.__name__
                else:  # str
                    package_name = package
                new_package_name = ".".join([package_name, resource_name])
                new_destination_path = pth.join(destination_path, resource_name)
                copy_resource_folder(new_package_name, new_destination_path, exclude=exclude)



"""
Computation of beam section that leads to yield stress.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om

import fastoad.api as oad


@oad.RegisterOpenMDAOSystem("tutorial.beam_problem.stresses")
class Stress(om.ExplicitComponent):
    """
    Computes needed beam section second moment of area that leads to yield stress, given force and
    material properties.
    """

    def setup(self):

        self.add_input("data:geometry:L", val=np.nan, units="m")
        self.add_input("data:geometry:h", val=np.nan, units="m")
        self.add_input("data:forces:F", val=np.nan, units="N")
        self.add_input("data:weight:linear_weight", val=np.nan, units="N/m")
        self.add_input("data:material:yield_stress", val=np.nan, units="Pa")

        self.add_output("data:geometry:Ixx", val=1e-5, units="m**4")

    def setup_partials(self):

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):

        L = inputs["data:geometry:L"]
        h = inputs["data:geometry:h"]
        F = inputs["data:forces:F"]
        w = inputs["data:weight:linear_weight"]
        s = inputs["data:material:yield_stress"]

        # Max bending location along the beam
        if w * L - F > 0:
            y_max = (w * L - F) / w
        else:
            y_max = 0
        outputs["data:geometry:Ixx"] = (L - y_max) * (F - 0.5 * w * (L - y_max)) * h / (2 * s)



"""Management of mission input definitions."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import InitVar, dataclass, field
from numbers import Number
from typing import Iterable, Mapping, Optional, Tuple, Union

import numpy as np
from openmdao import api as om

from fastoad.model_base import FlightPoint
from fastoad.openmdao.variables import Variable
from .constants import BASE_UNITS


@dataclass
class InputDefinition:
    """
    Class for managing definition of mission inputs.

    It stores and processes input definition from mission files:
        - provides values to be used for mission computation (management of units and variables)
        - provides information for OpenMDAO declaration
    """

    #: The parameter this input is defined for.
    parameter_name: str

    #: Value, matching `input_unit`. At instantiation, it can also be the variable name.
    input_value: Optional[Union[Number, Iterable, str]]

    #: Unit used for self.input_value.
    input_unit: Optional[str] = None

    #: Default value. Used if value is a variable name.
    default_value: Number = np.nan

    #: True if variable is defined as relative.
    is_relative: bool = False

    #: Prefix used when generating variable name because "~" was used in variable name input.
    part_identifier: str = ""

    #: Unit used for self.value. Automatically determined from self.parameter_name,
    #: mainly from unit definition for FlightPoint class.
    output_unit: Optional[str] = field(default=None, init=False, repr=False)

    #: Value of the "shape" openMDAO flag for input declaration.
    shape: Optional[Tuple[int]] = None

    #: Value of the "shape_by_conn" openMDAO flag for input declaration.
    shape_by_conn: bool = False

    #: Used only for tests
    variable_name: InitVar[Optional[str]] = None

    #: Used only for tests
    use_opposite: InitVar[Optional[bool]] = None

    #: True if the opposite value should be used, if input is defined by a variable.
    _use_opposite: bool = field(default=False, init=False, repr=True)

    _variable_name: Optional[str] = field(default=None, init=False, repr=True)

    def __post_init__(self, variable_name, use_opposite):

        if self.parameter_name.startswith("delta_"):
            self.is_relative = True
            self.parameter_name = self.parameter_name[6:]

        self.output_unit = FlightPoint.get_units().get(self.parameter_name)
        if self.output_unit is None:
            self.output_unit = BASE_UNITS.get(self.parameter_name)
        if self.output_unit == "-":
            self.output_unit = None

        if self.input_unit is None:
            self.input_unit = self.output_unit

        if variable_name and not isinstance(variable_name, property):
            # dataclass "feature": default value of 'variable_name' is 'property' because it
            # is defined as a property.
            self.variable_name = variable_name
            self.input_value = None
        elif isinstance(self.input_value, str) and (
            ":" in self.input_value or self.input_value.startswith(("~", "-~"))
        ):
            # This is done at end of initialization, because self.variable_name property may need
            # data as self.parameter_name, self.prefix...
            self.variable_name = self.input_value
            self.input_value = None

        if use_opposite is not None:
            self._use_opposite = use_opposite

    @property
    def value(self):
        """

        :return: Value of variable in DEFAULT unit (unit used by mission calculation),
                 or None if input is a variable and set_variable_input() has NOT been called,
                 or the unchanged value if it is not a number.
        """
        try:
            return om.convert_units(self.input_value, self.input_unit, self.output_unit)
        except TypeError:
            return self.input_value

    @classmethod
    def from_dict(cls, parameter_name, definition_dict: dict, part_identifier=None):
        """
        Instantiates InputDefinition from definition_dict.

        definition_dict["value"] is used as `input_value` in instantiation. It can be an actual
        value or a variable name.

        :param parameter_name: used if definition_dict["value"] == "~" (or "-~")
        :param definition_dict: dict with keys ("value", "unit", "default"). "unit" and "default"
                                are optional.
        :param part_identifier: used if "~" is in definition_dict["value"]
        """
        if "value" not in definition_dict:
            return None

        input_def = cls(
            parameter_name,
            definition_dict["value"],
            input_unit=definition_dict.get("unit"),
            default_value=definition_dict.get("default", np.nan),
            shape_by_conn=definition_dict.get("shape_by_conn", False),
            part_identifier=part_identifier,
        )
        return input_def

    def set_variable_value(self, inputs: Mapping):
        """
        Sets numerical value from OpenMDAO inputs.

        OpenMDAO value is assumed to be provided with unit self.input_unit.

        :param inputs:
        """
        if self.variable_name:
            # Note: OpenMDAO `inputs` object has no `get()` method, so we need to do this:
            value = (
                inputs[self.variable_name] if self.variable_name in inputs else self.default_value
            )
            if self._use_opposite:
                self.input_value = -value
            else:
                self.input_value = value

    def get_input_definition(self) -> Optional[Variable]:
        """
        Provides information for input definition in OpenMDAO.

        :return: Variable instance with input definition, or None if no variable name was defined.
        """
        if self.variable_name:
            return Variable(
                name=self.variable_name,
                val=self.default_value,
                shape_by_conn=self.shape_by_conn,
                units=self.input_unit,
                desc="Input defined by the mission.",
            )
        return None

    # pylint: disable=function-redefined  # the variable_name field is an InitVar.
    @property
    def variable_name(self):
        """Associated variable name."""
        return self._variable_name

    @variable_name.setter
    def variable_name(self, var_name: Optional[str]):
        if isinstance(var_name, str):
            self._use_opposite = var_name.startswith("-")
            var_name = var_name.strip("- ")

            if "~" in var_name:
                # We authorize colons next to "~", but we do as they were not present.
                var_name = var_name.replace(":~", "~").replace("~:", "~")
                parts = var_name.replace(":~", "~").split("~")
                if len(parts) > 2:
                    # Hidden feature for now: using a double tilda (~~) will trigger
                    # a replacement by the mission name only
                    # (useful for backward compatibility since we need to provide
                    #  "data:mission:<mission_name>:TOW" as variable
                    prefix = parts[0]
                    suffix = parts[-1]
                    part_identifier = self.part_identifier.split(":")[0]
                else:
                    # Standard case with simple tilda
                    prefix, suffix = var_name.replace(":~", "~").split("~")
                    part_identifier = self.part_identifier

                if not prefix:
                    # If nothing before "~", a default value is used
                    prefix = "data:mission"
                if not suffix:
                    # If nothing after "~", the parameter name is used
                    suffix = self.parameter_name

                var_name = ":".join([prefix, part_identifier, suffix])

            self._variable_name = var_name
        else:
            self._variable_name = None

    def __str__(self):
        return str(self.value)



"""Classes for acceleration/deceleration segments."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from dataclasses import dataclass
from typing import List, Tuple

from fastoad.model_base import FlightPoint
from .base import AbstractManualThrustSegment
from ..exceptions import FastFlightSegmentIncompleteFlightPoint

_LOGGER = logging.getLogger(__name__)  # Logger for this module


@dataclass
class SpeedChangeSegment(AbstractManualThrustSegment, mission_file_keyword="speed_change"):
    """
    Computes a flight path segment where speed is modified with no change in altitude.

    The target must define a speed value among true_airspeed, equivalent_airspeed
    and mach.
    """

    def get_distance_to_target(
        self, flight_points: List[FlightPoint], target: FlightPoint
    ) -> float:
        if target.true_airspeed is not None:
            return target.true_airspeed - flight_points[-1].true_airspeed
        if target.equivalent_airspeed is not None:
            return target.equivalent_airspeed - flight_points[-1].equivalent_airspeed
        if target.mach is not None:
            return target.mach - flight_points[-1].mach

        raise FastFlightSegmentIncompleteFlightPoint(
            "No valid target definition for altitude change."
        )

    def get_gamma_and_acceleration(self, flight_point: FlightPoint) -> Tuple[float, float]:
        acceleration = (flight_point.thrust - flight_point.drag) / flight_point.mass
        return 0.0, acceleration



"""
Schema for mission definition files.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
from collections import OrderedDict
from importlib.resources import open_text
from os import PathLike
from typing import Union

from ensure import Ensure
from jsonschema import validate
from ruamel.yaml import YAML

from . import resources

JSON_SCHEMA_NAME = "mission_schema.json"

# Tags
SEGMENT_TAG = "segment"
ROUTE_TAG = "route"
PHASE_TAG = "phase"
RESERVE_TAG = "reserve"
PARTS_TAG = "parts"
CLIMB_PARTS_TAG = "climb_parts"
CRUISE_PART_TAG = "cruise_part"
DESCENT_PARTS_TAG = "descent_parts"
MISSION_DEFINITION_TAG = "missions"
ROUTE_DEFINITIONS_TAG = "routes"
PHASE_DEFINITIONS_TAG = "phases"
POLAR_TAG = "polar"


class MissionDefinition(OrderedDict):
    def __init__(self, file_path: Union[str, PathLike] = None):
        """
        Class for reading a mission definition from a YAML file.

        Path of YAML file should be provided at instantiation, or in
        :meth:`load`.

        :param file_path: path of YAML file to read.
        """
        super().__init__()
        if file_path:
            self.load(file_path)

    def load(self, file_path: Union[str, PathLike]):
        """
        Loads a mission definition from provided file path.

        Any existing definition will be overwritten.

        :param file_path: path of YAML file to read.
        """
        self.clear()
        yaml = YAML(typ="safe", pure=True)

        with open(file_path) as yaml_file:
            data = yaml.load(yaml_file)

        with open_text(resources, JSON_SCHEMA_NAME) as json_file:
            json_schema = json.loads(json_file.read())
        validate(data, json_schema)

        self._validate(data)
        self.update(data)

    @classmethod
    def _validate(cls, content: dict):
        """
        Does a second pass validation of file content.

        Also applies thess features:
            * None values are set back to "~".
            *       - polar: foo:bar
                is translated to:
                    - polar:
                        CL: foo:bar:CL
                        CD: foo:bar:CD

        Errors are raised if file content is incorrect.

        :param content:
        """
        phase_names = set(content[PHASE_DEFINITIONS_TAG].keys())

        if ROUTE_DEFINITIONS_TAG in content:
            for route_definition in content[ROUTE_DEFINITIONS_TAG].values():
                for part in list(route_definition[CLIMB_PARTS_TAG]) + list(
                    route_definition[DESCENT_PARTS_TAG]
                ):
                    Ensure(part[PHASE_TAG]).is_in(phase_names)

        for mission_definition in content[MISSION_DEFINITION_TAG].values():
            reserve_count = 0
            for part in mission_definition[PARTS_TAG]:
                part_type, value = tuple(*part.items())
                if part_type == PHASE_TAG:
                    Ensure(value).is_in(content[PHASE_DEFINITIONS_TAG].keys())
                elif part_type == ROUTE_TAG:
                    Ensure(value).is_in(content[ROUTE_DEFINITIONS_TAG].keys())
                elif part_type == RESERVE_TAG:
                    reserve_count += 1
                    Ensure(value["ref"]).is_in(content[ROUTE_DEFINITIONS_TAG].keys())
            Ensure(reserve_count).is_less_than_or_equal_to(1)
            if reserve_count == 1:
                # reserve definition should be the last part
                Ensure(part_type).equals(RESERVE_TAG)

        cls._convert_none_values(content)

    @classmethod
    def _convert_none_values(cls, struct: Union[dict, list]):
        """
        Recursively transforms any None value in struct to "~"
        """
        if isinstance(struct, dict):
            for key, value in struct.items():
                if value is None:
                    struct[key] = "~"
                else:
                    cls._convert_none_values(value)
        elif isinstance(struct, list):
            for item in struct:
                cls._convert_none_values(item)



def update_mtow(owe, payload, mission_fuel):
    """
    Updates the MTOW based on the structural weight computed, the payload and the fuel consumed
    during the design mission.

    :param owe: the structural mass, in kg
    :param payload: the payload mass, in kg
    :param mission_fuel: the fuel consumed during the designated mission, in kg

    return mtow_new: the new Maximum Take-Off Weight based on the current iteration's computation,
    in kg
    """

    # Let's simply add the weight we computed
    mtow_new = owe + payload + mission_fuel

    return mtow_new



"""
Module for building OpenMDAO problem from configuration file
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import logging
import os.path as pth
from abc import ABC, abstractmethod
from importlib.resources import open_text
from typing import Dict

import openmdao.api as om
import tomlkit
from jsonschema import validate
from ruamel.yaml import YAML

from fastoad._utils.files import make_parent_dir
from fastoad.io import DataFile, IVariableIOFormatter
from fastoad.module_management.service_registry import RegisterOpenMDAOSystem, RegisterSubmodel
from fastoad.openmdao.problem import FASTOADProblem
from fastoad.openmdao.variables import VariableList
from . import resources
from .exceptions import (
    FASTConfigurationBadOpenMDAOInstructionError,
    FASTConfigurationBaseKeyBuildingError,
)

_LOGGER = logging.getLogger(__name__)  # Logger for this module

KEY_FOLDERS = "module_folders"
KEY_INPUT_FILE = "input_file"
KEY_OUTPUT_FILE = "output_file"
KEY_COMPONENT_ID = "id"
KEY_CONNECTION_ID = "connections"
KEY_MODEL = "model"
KEY_SUBMODELS = "submodels"
KEY_DRIVER = "driver"
KEY_OPTIMIZATION = "optimization"
KEY_DESIGN_VARIABLES = "design_variables"
KEY_CONSTRAINTS = "constraints"
KEY_OBJECTIVE = "objective"
JSON_SCHEMA_NAME = "configuration.json"


class FASTOADProblemConfigurator:
    """
    class for configuring an OpenMDAO problem from a configuration file

    See :ref:`description of configuration file <configuration-file>`.

    :param conf_file_path: if provided, configuration will be read directly from it
    """

    def __init__(self, conf_file_path=None):
        self._conf_file = None

        self._serializer = _YAMLSerializer()

        # self._configuration_modifier offers a way to modify problems after
        # they have been generated from configuration (private usage for now)
        self._configuration_modifier: "_IConfigurationModifier" = None

        if conf_file_path:
            self.load(conf_file_path)

    @property
    def input_file_path(self):
        """path of file with input variables of the problem"""
        path = str(self._serializer.data[KEY_INPUT_FILE])
        if not pth.isabs(path):
            path = pth.normpath(pth.join(pth.dirname(self._conf_file), path))
        return path

    @input_file_path.setter
    def input_file_path(self, file_path: str):
        self._serializer.data[KEY_INPUT_FILE] = file_path

    @property
    def output_file_path(self):
        """path of file where output variables will be written"""
        path = str(self._serializer.data[KEY_OUTPUT_FILE])
        if not pth.isabs(path):
            path = pth.normpath(pth.join(pth.dirname(self._conf_file), path))
        return path

    @output_file_path.setter
    def output_file_path(self, file_path: str):
        self._serializer.data[KEY_OUTPUT_FILE] = file_path

    def get_problem(self, read_inputs: bool = False, auto_scaling: bool = False) -> FASTOADProblem:
        """
        Builds the OpenMDAO problem from current configuration.

        :param read_inputs: if True, the created problem will already be fed
                            with variables from the input file
        :param auto_scaling: if True, automatic scaling is performed for design
                             variables and constraints
        :return: the problem instance
        """
        if self._serializer.data is None:
            raise RuntimeError("read configuration file first")

        problem = FASTOADProblem()
        self._build_model(problem)
        problem.input_file_path = self.input_file_path
        problem.output_file_path = self.output_file_path

        if read_inputs:
            problem.read_inputs()

        driver = self._serializer.data.get(KEY_DRIVER, "")
        if driver:
            problem.driver = _om_eval(driver)

        if self.get_optimization_definition():
            self._add_constraints(problem.model, auto_scaling)
            self._add_objectives(problem.model)

        if read_inputs:
            self._add_design_vars(problem.model, auto_scaling)

        if self._configuration_modifier:
            self._configuration_modifier.modify(problem)

        return problem

    def load(self, conf_file):
        """
        Reads the problem definition

        :param conf_file: Path to the file to open or a file descriptor
        """

        self._conf_file = pth.abspath(conf_file)  # for resolving relative paths
        conf_dirname = pth.dirname(self._conf_file)

        if pth.splitext(self._conf_file)[-1] == ".toml":
            self._serializer = _TOMLSerializer()
            _LOGGER.warning(
                "TOML-formatted configuration files are deprecated. Please use YAML format."
            )
        else:
            self._serializer = _YAMLSerializer()
        self._serializer.read(self._conf_file)

        # Syntax validation
        with open_text(resources, JSON_SCHEMA_NAME) as json_file:
            json_schema = json.loads(json_file.read())
        validate(self._serializer.data, json_schema)
        # Issue a simple warning for unknown keys at root level
        for key in self._serializer.data:
            if key not in json_schema["properties"].keys():
                _LOGGER.warning('Configuration file: "%s" is not a FAST-OAD key.', key)

        # Looking for modules to register
        module_folder_paths = self._serializer.data.get(KEY_FOLDERS)
        if isinstance(module_folder_paths, str):
            module_folder_paths = [module_folder_paths]
        if module_folder_paths:
            for folder_path in module_folder_paths:
                folder_path = pth.join(conf_dirname, str(folder_path))
                if not pth.exists(folder_path):
                    _LOGGER.warning("SKIPPED %s: it does not exist.", folder_path)
                else:
                    RegisterOpenMDAOSystem.explore_folder(folder_path)

        # Settings submodels
        RegisterSubmodel.cancel_submodel_deactivations()
        submodel_specs = self._serializer.data.get(KEY_SUBMODELS, {})
        for submodel_requirement, submodel_id in submodel_specs.items():
            RegisterSubmodel.active_models[submodel_requirement] = submodel_id

    def save(self, filename: str = None):
        """
        Saves the current configuration
        If no filename is provided, the initially read file is used.

        :param filename: file where to save configuration
        """
        if not filename:
            filename = self._conf_file

        make_parent_dir(filename)
        self._serializer.write(filename)

    def write_needed_inputs(
        self, source_file_path: str = None, source_formatter: IVariableIOFormatter = None
    ):
        """
        Writes the input file of the problem with unconnected inputs of the
        configured problem.

        Written value of each variable will be taken:

            1. from input_data if it contains the variable
            2. from defined default values in component definitions

        :param source_file_path: if provided, variable values will be read from it
        :param source_formatter: the class that defines format of input file. if
                                 not provided, expected format will be the default one.
        """
        problem = self.get_problem(read_inputs=False)
        problem.setup()
        variables = DataFile(self.input_file_path, load_data=False)

        unconnected_inputs = VariableList.from_problem(
            problem,
            use_initial_values=True,
            get_promoted_names=True,
            promoted_only=True,
            io_status="inputs",
        )

        variables.update(
            unconnected_inputs,
            add_variables=True,
        )
        if source_file_path:
            ref_vars = DataFile(source_file_path, formatter=source_formatter)
            variables.update(ref_vars, add_variables=False)
            for var in variables:
                var.is_input = True
        variables.save()

    def get_optimization_definition(self) -> Dict:
        """
        Returns information related to the optimization problem:
            - Design Variables
            - Constraints
            - Objectives

        :return: dict containing optimization settings for current problem
        """

        optimization_definition = {}
        conf_dict = self._serializer.data.get(KEY_OPTIMIZATION)
        if conf_dict:
            for sec, elements in conf_dict.items():
                optimization_definition[sec] = {elem["name"]: elem for elem in elements}
        return optimization_definition

    def set_optimization_definition(self, optimization_definition: Dict):
        """
        Updates configuration with the list of design variables, constraints, objectives
        contained in the optimization_definition dictionary.

        Keys of the dictionary are: "design_var", "constraint", "objective".

        Configuration file will not be modified until :meth:`save` is used.

        :param optimization_definition: dict containing the optimization problem definition
        """
        subpart = {}
        for key, value in optimization_definition.items():
            subpart[key] = [value for _, value in optimization_definition[key].items()]
        subpart = {"optimization": subpart}
        self._serializer.data.update(subpart)

    def _build_model(self, problem: FASTOADProblem):
        """
        Builds the problem model as defined in the configuration file.

        The problem model is populated with subsystems indicated in configuration file.
        """

        model = problem.model
        model.active_submodels = self._serializer.data.get(KEY_SUBMODELS, {})

        model_definition = self._serializer.data.get(KEY_MODEL)

        try:
            if KEY_COMPONENT_ID in model_definition:
                # The defined model is only one system
                system_id = model_definition[KEY_COMPONENT_ID]
                sub_component = RegisterOpenMDAOSystem.get_system(system_id)
                model.add_subsystem("system", sub_component, promotes=["*"])
            else:
                # The defined model is a group
                self._parse_problem_table(model, model_definition)

        except FASTConfigurationBaseKeyBuildingError as err:
            log_err = err.__class__(err, KEY_MODEL)
            _LOGGER.error(log_err)
            raise log_err

    def _parse_problem_table(self, group: om.Group, table: dict):
        """
        Feeds provided *group*, using definition in provided TOML *table*.

        :param group:
        :param table:
        """
        # assert isinstance(table, dict), "table should be a dictionary"

        for key, value in table.items():
            if isinstance(value, dict):  # value defines a sub-component
                if KEY_COMPONENT_ID in value:
                    # It is a non-group component, that should be registered with its ID
                    options = value.copy()
                    identifier = options.pop(KEY_COMPONENT_ID)

                    # Process option values that are relative paths
                    conf_dirname = pth.dirname(self._conf_file)
                    for name, option_value in options.items():
                        option_is_path = (
                            name.endswith("file")
                            or name.endswith("path")
                            or name.endswith("dir")
                            or name.endswith("directory")
                            or name.endswith("folder")
                        )
                        if (
                            isinstance(option_value, str)
                            and option_is_path
                            and not pth.isabs(option_value)
                        ):
                            options[name] = pth.join(conf_dirname, option_value)

                    sub_component = RegisterOpenMDAOSystem.get_system(identifier, options=options)
                    group.add_subsystem(key, sub_component, promotes=["*"])
                else:
                    # It is a Group
                    sub_component = group.add_subsystem(key, om.Group(), promotes=["*"])
                    try:
                        self._parse_problem_table(sub_component, value)
                    except FASTConfigurationBadOpenMDAOInstructionError as err:
                        # There has been an error while parsing an attribute.
                        # Error is relayed with key added for context
                        raise FASTConfigurationBadOpenMDAOInstructionError(err, key)
            elif key == KEY_CONNECTION_ID and isinstance(value, list):
                # a list of dict currently defines only connections
                for connection_def in value:
                    group.connect(connection_def["source"], connection_def["target"])
            else:
                # value is an attribute of current component and will be literally interpreted
                try:
                    setattr(group, key, _om_eval(str(value)))  # pylint:disable=eval-used
                except Exception as err:
                    raise FASTConfigurationBadOpenMDAOInstructionError(err, key, value)

    def _add_constraints(self, model, auto_scaling):
        """
        Adds constraints to provided model as instructed in current configuration

        :param model:
        :param auto_scaling:
        :return:
        """
        optimization_definition = self.get_optimization_definition()
        # Constraints
        constraint_tables = optimization_definition.get(KEY_CONSTRAINTS, {})
        for constraint_table in constraint_tables.values():
            if (
                auto_scaling
                and "lower" in constraint_table
                and "upper" in constraint_table
                and constraint_table.get("ref0") is not None
                and constraint_table.get("ref") is not None
                and constraint_table["lower"] != constraint_table["upper"]
            ):
                constraint_table["ref0"] = constraint_table["lower"]
                constraint_table["ref"] = constraint_table["upper"]
            model.add_constraint(**constraint_table)

    def _add_objectives(self, model):
        """
        Adds objectives to provided model as instructed in current configuration

        :param model:
        :return:
        """
        optimization_definition = self.get_optimization_definition()
        objective_tables = optimization_definition.get(KEY_OBJECTIVE, {})
        for objective_table in objective_tables.values():
            model.add_objective(**objective_table)

    def _add_design_vars(self, model, auto_scaling):
        """
        Adds design variables to provided model as instructed in current configuration

        :param model:
        :param auto_scaling:
        :return:
        """
        optimization_definition = self.get_optimization_definition()
        design_var_tables = optimization_definition.get(KEY_DESIGN_VARIABLES, {})
        for design_var_table in design_var_tables.values():
            if (
                auto_scaling
                and "lower" in design_var_table
                and "upper" in design_var_table
                and design_var_table.get("ref0") is not None
                and design_var_table.get("ref") is not None
                and design_var_table["lower"] != design_var_table["upper"]
            ):
                design_var_table["ref0"] = design_var_table["lower"]
                design_var_table["ref"] = design_var_table["upper"]
            model.add_design_var(**design_var_table)

    def _set_configuration_modifier(self, modifier: "_IConfigurationModifier"):
        self._configuration_modifier = modifier


def _om_eval(string_to_eval: str):
    """
    Evaluates strings that assume `import openmdao.api as om` is done.

    eval() is used for that, as safely as possible.

    :param string_to_eval:
    :return: result of eval()
    """
    if "__" in string_to_eval:
        raise ValueError("No double underscore allowed in evaluated string for security reasons")
    return eval(string_to_eval, {"__builtins__": {}}, {"om": om})


class _IDictSerializer(ABC):
    """Interface for reading and writing dict-like data"""

    @property
    @abstractmethod
    def data(self) -> dict:
        """
        The data that have been read, or will be written.
        """

    @abstractmethod
    def read(self, file_path: str):
        """
        Reads data from provided file.
        :param file_path:
        """

    @abstractmethod
    def write(self, file_path: str):
        """
        Writes data to provided file.
        :param file_path:
        """


class _TOMLSerializer(_IDictSerializer):
    """TOML-format serializer."""

    def __init__(self):
        self._data = None

    @property
    def data(self):
        return self._data

    def read(self, file_path: str):
        with open(file_path, "r") as toml_file:
            self._data = tomlkit.loads(toml_file.read())

    def write(self, file_path: str):
        with open(file_path, "w") as file:
            file.write(tomlkit.dumps(self._data))


class _YAMLSerializer(_IDictSerializer):
    """YAML-format serializer."""

    def __init__(self):
        self._data = None

    @property
    def data(self):
        return self._data

    def read(self, file_path: str):
        yaml = YAML(typ="safe")
        with open(file_path) as yaml_file:
            self._data = yaml.load(yaml_file)

    def write(self, file_path: str):
        yaml = YAML()
        yaml.default_flow_style = False
        with open(file_path, "w") as file:
            yaml.dump(self._data, file)


class _IConfigurationModifier(ABC):
    """
    Interface for a configuration modifier used in FASTOADProblemConfigurator.
    """

    @abstractmethod
    def modify(self, problem: om.Problem):
        """
        This method will do operations on the provided problem.

        problem.setup() is assumed NOT called.
        """



import openmdao.api as om

from .sub_components.compute_profile_drag import ComputeProfileDrag
from .sub_components.compute_induced_drag_coefficient import ComputeInducedDragCoefficient
from .sub_components.compute_lift_to_drag_ratio import ComputeLiftToDragRatio


class ComputeAerodynamics(om.Group):
    def setup(self):

        self.add_subsystem(name="compute_profile_drag", subsys=ComputeProfileDrag(), promotes=["*"])
        self.add_subsystem(
            name="compute_induced_drag", subsys=ComputeInducedDragCoefficient(), promotes=["*"]
        )
        self.add_subsystem(
            name="compute_lift_to_drag", subsys=ComputeLiftToDragRatio(), promotes=["*"]
        )



import scipy.constants as sc

from fastoad.model_base.atmosphere import Atmosphere


def compute_l_d(cruise_altitude, cruise_speed, cd0, k, mtow, wing_area):
    """
    Computes the lift to drag ratio considering a lift equilibrium in cruise and a simple quadratic model

    :param cruise_altitude: Cruise altitude, in m
    :param cruise_speed: Cruise speed, in m/s
    :param cd0: Profile drag, no unit
    :param k: Lift induced drag coefficient, no unit
    :param mtow: Max Take-Off Weight, in kg
    :param wing_area: Wing area, in m2

    :return l_d: Lift-to-drag ratio in cruise conditions, no unit
    """

    # Air density at cruise level, to compute it, we will use the Atmosphere model available in
    # FAST-OAD, so we will create an Atmosphere instance using the cruise altitude and extract
    # its density attribute
    atm = Atmosphere(altitude=cruise_altitude, altitude_in_feet=False)
    rho = atm.density

    # Computation of the cruise lift coefficient using a simple equilibrium
    cl = (mtow * sc.g) / (0.5 * rho * cruise_speed ** 2.0 * wing_area)

    # Computation of the cruise drag coefficient using the simple quadratic model
    cd = cd0 + k * cl ** 2

    # Computation of the ratio
    l_d = cl / cd

    return l_d



"""Definition of globally used constants."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from aenum import Enum, IntEnum, extend_enum


class FlightPhase(Enum):
    """
    Enumeration of flight phases.
    """

    TAXI_OUT = "taxi_out"
    TAKEOFF = "takeoff"
    INITIAL_CLIMB = "initial_climb"
    CLIMB = "climb"
    CRUISE = "cruise"
    DESCENT = "descent"
    LANDING = "landing"
    TAXI_IN = "taxi_in"


class EngineSetting(IntEnum):
    """
    Enumeration of engine settings.
    """

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name.lower() == other.lower()

        return super().__eq__(other)

    def __hash__(self):
        return self.value

    @classmethod
    def convert(cls, name: str) -> "EngineSetting":
        """
        :param name:
        :return: the EngineSetting instance that matches the provided name (case-insensitive)
        """
        for instance in cls:
            if instance.name.lower() == name.lower():
                return instance

        return None


# Using the extensibility of EngineSetting is not needed here, but it allows to
# test it.
extend_enum(EngineSetting, "TAKEOFF")
extend_enum(EngineSetting, "CLIMB")
extend_enum(EngineSetting, "CRUISE")
extend_enum(EngineSetting, "IDLE")


class RangeCategory(Enum):
    """
    Definition of lower and upper limits of aircraft range categories, in Nautical Miles.

    can be used like::
        >>> range_value = 800.
        >>> range_value in RangeCategory.SHORT
        True

    which is equivalent to:
        >>>  RangeCategory.SHORT.min() <= range_value <= RangeCategory.SHORT.max()
    """

    SHORT = (0.0, 1500.0)
    SHORT_MEDIUM = (1500.0, 3000.0)
    MEDIUM = (3000.0, 4500.0)
    LONG = (4500.0, 6000.0)
    VERY_LONG = (6000.0, 1.0e6)

    def min(self):
        """
        :return: minimum range in category
        """
        return self.value[0]

    def max(self):
        """
        :return: maximum range in category
        """
        return self.value[1]

    def __contains__(self, range_value):
        """
        :param range_value:
        :return: True if rang_value is inside range limits, False otherwise
        """
        return self.min() <= range_value <= self.max()



"""
Enables to remove duplicate variables from text file
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from os import remove


def remove_duplicate(input_file_name: str):
    """
    Removes duplicate lines in a text file

    """
    input_file = open(input_file_name, "r")

    # Store all the variables
    stored_lines = []
    for line in input_file:
        if line not in stored_lines:
            stored_lines.append(line)
    # Delete the file
    input_file.close()
    remove(input_file_name)

    # Create new file
    input_file = open(input_file_name, "w")

    # Write non duplicate lines
    for line in stored_lines:
        input_file.write(str(line))

    input_file.close()


if __name__ == "__main__":

    remove_duplicate("legacy1.txt")



#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fnmatch import fnmatchcase
from os.path import exists, isfile
from typing import IO, List, Sequence, Union

from fastoad.openmdao.variables import VariableList
from . import IVariableIOFormatter
from .xml import VariableXmlStandardFormatter
from ..exceptions import FastError


class VariableIO:
    """
    Class for reading and writing variable values from/to file.

    The file format is defined by the class provided as `formatter` argument.

    :param data_source: the I/O stream, or a file path, used for reading or writing data
    :param formatter: a class that determines the file format to be used. Defaults to a
                      VariableBasicXmlFormatter instance.
    """

    def __init__(self, data_source: Union[str, IO], formatter: IVariableIOFormatter = None):
        self.data_source = data_source
        self.formatter: IVariableIOFormatter = (
            formatter if formatter else VariableXmlStandardFormatter()
        )

    def read(self, only: List[str] = None, ignore: List[str] = None) -> VariableList:
        """
        Reads variables from provided data source.

        Elements of `only` and `ignore` can be real variable names or Unix-shell-style patterns.
        In any case, comparison is case-sensitive.

        :param only: List of variable names that should be read. Other names will be
                     ignored. If None, all variables will be read.
        :param ignore: List of variable names that should be ignored when reading.
        :return: an VariableList instance where outputs have been defined using provided source
        """
        if isinstance(self.data_source, str) and not isfile(self.data_source):
            raise FileNotFoundError(
                f'File "{self.data_source}" is unavailable for reading.'
            ) from FastError()
        variables = self.formatter.read_variables(self.data_source)
        used_variables = self._filter_variables(variables, only=only, ignore=ignore)
        return used_variables

    def write(self, variables: VariableList, only: List[str] = None, ignore: List[str] = None):
        """
        Writes variables from provided VariableList instance.

        Elements of `only` and `ignore` can be real variable names or Unix-shell-style patterns.
        In any case, comparison is case-sensitive.

        :param variables: a VariableList instance
        :param only: List of variable names that should be written. Other names will be
                     ignored. If None, all variables will be written.
        :param ignore: List of variable names that should be ignored when writing
        """
        used_variables = self._filter_variables(variables, only=only, ignore=ignore)

        # Before writing, variables are sorted to have short paths first. With equal path length
        # alphanumeric order will be used.
        used_variables.sort(key=lambda var: "%02i_%s" % (len(var.name.split(":")), var.name))

        self.formatter.write_variables(self.data_source, used_variables)

    @staticmethod
    def _filter_variables(
        variables: VariableList, only: Sequence[str] = None, ignore: Sequence[str] = None
    ) -> VariableList:
        """
        filters the variables such that the ones in arg only are kept and the ones in
        arg ignore are removed.

        Elements of `only` and `ignore` can be variable names or Unix-shell-style patterns.
        In any case, filter is case-sensitive.

        :param variables:
        :param only: List of OpenMDAO variable names that should be written. Other names will be
                     ignored. If None, all variables will be written.
        :param ignore: List of OpenMDAO variable names that should be ignored when writing
        :return: filtered variables
        """

        # Dev note: We use sets, but sets of Variable instances do
        # not work. Do we work with variable names instead.
        # FIXME: Variable instances are now hashable, so set of Variable instances should now work

        var_names = variables.names()

        if only is None:
            used_var_names = set(var_names)
        else:
            used_var_names = set()
            for pattern in only:
                used_var_names.update(
                    [variable.name for variable in variables if fnmatchcase(variable.name, pattern)]
                )

        if ignore is not None:
            for pattern in ignore:
                used_var_names.difference_update(
                    [variable.name for variable in variables if fnmatchcase(variable.name, pattern)]
                )

        # It could be simpler, but I want to keep the order
        used_variables = VariableList()
        for var in variables:
            if var.name in used_var_names:
                used_variables.append(var)
        return used_variables


class DataFile(VariableList):
    """
    Class for managing FAST-OAD data files.

    Behaves like :class:`~fastoad.openmdao.variables.VariableList` class but has :meth:`load` and
    :meth:`save` methods.
    """

    def __init__(
        self,
        data_source: Union[str, IO, list] = None,
        formatter: IVariableIOFormatter = None,
        load_data=True,
    ):
        """
        If variable list is specified for data_source, :attr:`file_path` will have to be set before
        using :method:`save`.

        :param data_source: Can be the file path where data will be loaded and saved, or a list of
                            :class:`~fastoad.openmdao.variables.Variable` instances that will
                            be used for initialization (or a
                            :class:`~fastoad.openmdao.variables.VariableList` instance).
        :param formatter: (ignored if data_source is not an I/O stream nor a file path)
                          a class that determines the file format to be used. Defaults to FAST-OAD
                          native format. See :class:`VariableIO` for more information.
        :param load_data: (ignored if data_source is not an I/O stream nor a file path)
                          if True, file is expected to exist and its content will be loaded at
                          instantiation.
        """
        super().__init__()
        self._variable_io = VariableIO(None, formatter)
        if isinstance(data_source, (str, IO)):
            self.file_path = data_source
            if load_data:
                self.load()
        if isinstance(data_source, list):
            self.update(data_source)

    @property
    def file_path(self) -> str:
        """Path of data file."""
        return self._variable_io.data_source

    @file_path.setter
    def file_path(self, value: str):
        self._variable_io.data_source = value

    @property
    def formatter(self) -> IVariableIOFormatter:
        """Class that defines the file format."""
        return self._variable_io.formatter

    @formatter.setter
    def formatter(self, value: IVariableIOFormatter):
        self._variable_io.formatter = value

    def load(self):
        """Loads file content."""
        self.clear()
        self.update(self._variable_io.read(), add_variables=True)

    def save(self):
        """Saves current state of variables in file."""
        if self.file_path is None:
            raise FileNotFoundError(
                "Destination file not set. Please use .save_as() instead."
            ) from FastError()
        self._variable_io.write(self)

    def save_as(self, file_path: str, overwrite=False, formatter: IVariableIOFormatter = None):
        """
        Sets the associated file path as specified and saves current state of variables.

        :param file_path:
        :param overwrite: if specified file already exists and overwrite is False, an error is
                          triggered.
        :param formatter: a class that determines the file format to be used. Defaults to FAST-OAD
                          native format. See :class:`VariableIO` for more information.
        """
        if not overwrite and exists(file_path):
            raise FileExistsError(f'File "{file_path}" already exists.') from FastError()
        if formatter:
            self.formatter = formatter
        self.file_path = file_path
        self.save()



"""
Plugin system for declaration of FAST-OAD models.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os.path as pth
import sys
import warnings
from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, Dict, List

import numpy as np

from fastoad.openmdao.variables import Variable
from ._bundle_loader import BundleLoader
from .exceptions import (
    FastNoAvailableConfigurationFileError,
    FastNoDistPluginError,
    FastSeveralConfigurationFilesError,
    FastSeveralDistPluginsError,
    FastUnknownConfigurationFileError,
    FastUnknownDistPluginError,
)
from .._utils.resource_management.contents import PackageReader

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

_LOGGER = logging.getLogger(__name__)  # Logger for this module

OLD_MODEL_PLUGIN_ID = "fastoad_model"
MODEL_PLUGIN_ID = "fastoad.plugins"


class SubPackageNames(Enum):
    """Enumeration of possible plugin subpackages."""

    MODELS = "models"
    NOTEBOOKS = "notebooks"
    CONFIGURATIONS = "configurations"


@dataclass
class ResourceInfo:
    """Class for storing information about configuration files provided by plugins."""

    name: str
    dist_name: str
    plugin_name: str
    package_name: str


@dataclass
class PluginDefinition:
    """
    Stores and provides FAST-OAD plugin data.
    """

    dist_name: str
    plugin_name: str
    package_name: str = ""
    subpackages: Dict[SubPackageNames, str] = field(default_factory=dict)

    def detect_subfolders(self):
        """
        Scans plugin folders and populates :attr:`subpackages`.
        """
        package = PackageReader(self.package_name)
        for subpackage_name in SubPackageNames:
            if subpackage_name.value in package.contents:
                self.subpackages[subpackage_name] = ".".join(
                    [self.package_name, subpackage_name.value]
                )

    def get_configuration_file_list(self) -> List[ResourceInfo]:
        """
        :return: List of configuration files that are provided by the distribution.
        """
        if SubPackageNames.CONFIGURATIONS in self.subpackages:
            return [
                ResourceInfo(
                    name=file,
                    dist_name=self.dist_name,
                    plugin_name=self.plugin_name,
                    package_name=self.subpackages[SubPackageNames.CONFIGURATIONS],
                )
                for file in PackageReader(self.subpackages[SubPackageNames.CONFIGURATIONS]).contents
                if pth.splitext(file)[1] in [".yml", ".yaml"]
            ]

        return []


@dataclass()
class DistributionPluginDefinition(dict):
    """
    Stores and provides data for FAST-OAD plugins provided by a Python distribution.
    """

    dist_name: str = None

    def read_entry_point(self, entry_point: importlib_metadata.EntryPoint, group: str):
        """
        Adds plugin definition from provided entry point to

        Does nothing if entry_point.dist.project_name is not equal to :attr:`dist_name`

        :param entry_point:
        :param group:
        """
        if self.dist_name != entry_point.dist.name:
            return

        plugin_definition = PluginDefinition(
            dist_name=self.dist_name,
            plugin_name=entry_point.name,
        )
        plugin_definition.package_name = entry_point.module
        self[entry_point.name] = plugin_definition

        if group == OLD_MODEL_PLUGIN_ID:
            warnings.warn(
                f'"{self.dist_name}" package uses `fastoad_model` as plugin group ID, which is '
                "deprecated. `fastoad.plugins` should be used instead.",
                DeprecationWarning,
            )
            self[entry_point.name].subpackages[SubPackageNames.MODELS] = entry_point.module

        if group == MODEL_PLUGIN_ID:
            self[entry_point.name].detect_subfolders()

    def get_configuration_file_list(self, plugin_name=None) -> List[ResourceInfo]:
        """
        :param plugin_name: If provided, only file names provided by the plugin in
                            the distribution will be returned, or an empty list if
                            the plugin is not in the distribution.
        :return:  List of configuration file information that are provided by the distribution.
        """
        file_list = []
        if plugin_name:
            if plugin_name in self:
                file_list = self[plugin_name].get_configuration_file_list()
        else:
            for plugin in self.values():
                file_list += plugin.get_configuration_file_list()

        return file_list

    def get_configuration_file_info(self, file_name=None, plugin_name=None) -> ResourceInfo:
        """
        :param file_name: can be None if only one configuration file is provided in the
                          distribution (or in the plugin if `plugin_name` is provided)
        :param plugin_name:
        :return: information for specified configuration file name.
        :raise FastSeveralConfigurationFilesError: if several configuration files are available
                                                   but `file_name` has not been provided.
        :raise FastUnknownConfigurationFileError: if the specified configuration file is not
                                                  available.
        """
        conf_file_list = self.get_configuration_file_list(plugin_name)
        if not conf_file_list:
            raise FastNoAvailableConfigurationFileError()

        if file_name is None:
            if len(conf_file_list) > 1:
                raise FastSeveralConfigurationFilesError(self.dist_name)
            file_info = conf_file_list[0]
        else:
            matching_list = list(filter(lambda item: item.name == file_name, conf_file_list))
            if len(matching_list) == 0:
                raise FastUnknownConfigurationFileError(file_name, self.dist_name)

            # Here we implicitly assume that plugin developers will ensure that there will be
            # no duplicates in conf file names (possible if several plugins are in the
            # same installed package, but such practice is discouraged).
            file_info = matching_list[0]

        return file_info

    def get_notebook_folder_list(self, plugin_name=None) -> List[ResourceInfo]:
        """
        :param plugin_name: If provided, only notebook folder provided by the plugin (if any)
                            will be returned, or an empty list if the plugin is not in the
                            distribution.
        :return:  List of notebook folder information that are provided by the distribution.
        """
        if plugin_name:
            plugin_names = [plugin_name] if plugin_name in self else []
        else:
            plugin_names = self.keys()

        info = []
        for name in plugin_names:
            plugin_def = self[name]
            if SubPackageNames.NOTEBOOKS in plugin_def.subpackages:
                info.append(
                    ResourceInfo(
                        name=SubPackageNames.NOTEBOOKS.value,
                        dist_name=self.dist_name,
                        plugin_name=name,
                        package_name=plugin_def.subpackages[SubPackageNames.NOTEBOOKS],
                    )
                )
        return info

    def to_dict(self) -> Dict:
        """
        :return: a dict that contains plugin information
        """
        has_models = np.any(
            [SubPackageNames.MODELS in definition.subpackages for definition in self.values()]
        )
        has_notebooks = np.any(
            [SubPackageNames.NOTEBOOKS in definition.subpackages for definition in self.values()]
        )
        conf_files = [item.name for item in self.get_configuration_file_list()]
        return dict(
            installed_package=self.dist_name,
            has_models=has_models,
            has_notebooks=has_notebooks,
            configurations=sorted(conf_files),
        )


class FastoadLoader(BundleLoader):
    """
    Specialized :class:`BundleLoader` that will load plugins at first instantiation.

    I also provides data from available plugins
    """

    # This class attribute is private and is accessed through a property to ensure
    # that the class has been instantiated before the attribute is used.
    _dist_plugin_definitions: Dict[str, DistributionPluginDefinition]

    _loaded = False

    def __init__(self):
        super().__init__()
        if not self.__class__._loaded:
            # Setting cls.loaded to True already ensures that a second instantiation
            # during loading will not result in an import cycle.
            self.__class__._loaded = True
            self.__class__._dist_plugin_definitions = {}
            self.read_entry_points()
            self.load()

    @property
    def distribution_plugin_definitions(self) -> Dict[str, DistributionPluginDefinition]:
        """Stores plugin definitions with distribution names as dict keys."""
        return self._dist_plugin_definitions.copy()

    def get_distribution_plugin_definition(
        self, dist_name: str = None
    ) -> DistributionPluginDefinition:
        """
        :param dist_name: needed if more than one distribution with FAST-OAD plugin is installed.
        :return: the DistributionPluginDefinition instance that matches `dist_name`, if any
        :raise FastNoDistPluginError: if no distribution with plugin is available.
        :raise FastSeveralDistPluginsError: if several distributions are available but `dist_name`
                                        has not been provided.
        :raise FastUnknownDistPluginError: if the specified distribution is not available.
        """

        if len(self._dist_plugin_definitions) == 0:
            raise FastNoDistPluginError()

        if dist_name is None:
            if len(self._dist_plugin_definitions) > 1:
                raise FastSeveralDistPluginsError()
            return next(iter(self._dist_plugin_definitions.values()))

        if dist_name not in self._dist_plugin_definitions:
            raise FastUnknownDistPluginError(dist_name)

        return self._dist_plugin_definitions[dist_name]

    def get_configuration_file_list(self, dist_name: str) -> List[ResourceInfo]:
        """
        :param dist_name: the distribution to inspect
        :return: list of configuration files available for named distribution,
                 or an empty list if the specified distribution is not available
        """
        return self._get_resource_list(
            DistributionPluginDefinition.get_configuration_file_list,
            dist_name,
        )

    def get_notebook_folder_list(self, dist_name: str = None) -> List[ResourceInfo]:
        """
        Returns the list of notebook folders available for named distribution
        and optionally the named plugin of this distribution.

        :param dist_name: the distribution to inspect
        :return: list of notebook folders available for named distribution,
                 or an empty list if the specified distribution is not available
        """
        return self._get_resource_list(
            DistributionPluginDefinition.get_notebook_folder_list,
            dist_name,
        )

    def _get_resource_list(self, method: Callable, dist_name: str = None) -> List[ResourceInfo]:
        infos = []
        if dist_name:
            if dist_name not in self._dist_plugin_definitions:
                raise FastUnknownDistPluginError(dist_name)
            dist_names = [dist_name]
        else:
            if len(self._dist_plugin_definitions) == 0:
                raise FastNoDistPluginError()
            dist_names = self._dist_plugin_definitions.keys()

        for name in dist_names:
            dist_plugin_definitions = self._dist_plugin_definitions[name]
            infos += method(dist_plugin_definitions)

        return infos

    @classmethod
    def read_entry_points(cls):
        """
        Reads definitions of declared plugins.
        """
        # Dev note: in src/conftest.py, wrapt is used to monkey patch
        # importlib_metadata.entry_points when unit testing plugin-related feature.
        # This will not work if `from importlib_metadata import entry_points`
        # is done at beginning of the file, because then, we will use a reference
        # to original `entry_points`, before it is patched by conftest.py

        for group in [OLD_MODEL_PLUGIN_ID, MODEL_PLUGIN_ID]:
            for entry_point in importlib_metadata.entry_points(group=group):
                if entry_point.dist.name not in cls._dist_plugin_definitions:
                    cls._dist_plugin_definitions[
                        entry_point.dist.name
                    ] = DistributionPluginDefinition()
                plugin_dist = cls._dist_plugin_definitions[entry_point.dist.name]
                plugin_dist.dist_name = entry_point.dist.name
                plugin_dist.read_entry_point(entry_point, group)

    @classmethod
    def load(cls):
        """
        Loads declared plugins.
        """
        for dist_plugin_definitions in cls._dist_plugin_definitions.values():
            for plugin_name, plugin_def in dist_plugin_definitions.items():
                _LOGGER.info("Loading FAST-OAD plugin %s", plugin_name)
                cls._load_models(plugin_def)

    @classmethod
    def _load_models(cls, plugin_definition: PluginDefinition):
        """Loads models from plugin."""
        if SubPackageNames.MODELS in plugin_definition.subpackages:
            _LOGGER.debug("   Loading models")
            BundleLoader().explore_folder(
                plugin_definition.subpackages[SubPackageNames.MODELS], is_package=True
            )
            Variable.read_variable_descriptions(
                plugin_definition.subpackages[SubPackageNames.MODELS]
            )



"""
Exceptions for package configuration
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from fastoad.exceptions import FastError


class FASTConfigurationBaseKeyBuildingError(FastError):
    """
    Class for being raised from bottom to top of TOML dict so that in the end, the message
    provides the full qualified name of the problematic key.

    using `new_err = FASTConfigurationBaseKeyBuildingError(err, 'new_err_key', <value>)`:

    - if err is a FASTConfigurationBaseKeyBuildingError instance with err.key=='err_key':
        - new_err.key will be 'new_err_key.err_key'
        - new_err.value will be err.value (no need to provide a value here)
        - new_err.original_exception will be err.original_exception
    - otherwise, new_err.key will be 'new_err_key' and new_err.value will be <value>
        - new_err.key will be 'new_err_key'
        - new_err.value will be <value>
        - new_err.original_exception will be err

    :param original_exception: the error that happened for raising this one
    :param key: the current key
    :param value: the current value
    """

    def __init__(self, original_exception: Exception, key: str, value=None):
        """Constructor"""

        self.key = None
        """
        the "qualified key" (like "problem.group.component1") related to error, build
        through raising up the error
        """

        self.value = None
        """ the value related to error """

        self.original_exception = None
        """ the original error, when eval failed """

        if hasattr(original_exception, "key"):
            self.key = "%s.%s" % (key, original_exception.key)
        else:
            self.key = key
        if hasattr(original_exception, "value"):
            self.value = "%s.%s" % (value, original_exception.value)
        else:
            self.value = value
        if hasattr(original_exception, "original_exception"):
            self.original_exception = original_exception.original_exception
        else:
            self.original_exception = original_exception
        super().__init__(
            self,
            'Attribute or value not recognized : %s = "%s"\nOriginal error: %s'
            % (self.key, self.value, self.original_exception),
        )


class FASTConfigurationBadOpenMDAOInstructionError(FASTConfigurationBaseKeyBuildingError):
    """Class for managing errors that result from trying to set an attribute by eval."""



import openmdao.api as om

from ...OpenMDAO.mass.sub_components.compute_wing_mass import ComputeWingMass
from ...OpenMDAO.mass.sub_components.compute_owe import ComputeOwe

import fastoad.api as oad


@oad.RegisterOpenMDAOSystem("tutorial.fast_oad.mass")
class ComputeMass(om.Group):
    def setup(self):

        self.add_subsystem(name="compute_wing_mass", subsys=ComputeWingMass(), promotes=["*"])
        self.add_subsystem(name="compute_owe", subsys=ComputeOwe(), promotes=["*"])



#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Union, IO

from fastoad.openmdao.variables import VariableList


class IVariableIOFormatter(ABC):
    """
    Interface for formatter classes to be used in VariableIO class.

    The file format is defined by the implementation of this interface.
    """

    @abstractmethod
    def read_variables(self, data_source: Union[str, IO]) -> VariableList:
        """
        Reads variables from provided data source file.

        :param data_source:
        :return: a list of Variable instance
        """

    @abstractmethod
    def write_variables(self, data_source: Union[str, IO], variables: VariableList):
        """
        Writes variables to defined data source file.

        :param data_source:
        :param variables:
        """



"""
Class for managing an OpenMDAO variable.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os.path as pth
from importlib.resources import open_text
from typing import Dict, Hashable, Iterable, Mapping, Tuple, Union

import numpy as np
import openmdao.api as om

from fastoad._utils.resource_management.contents import PackageReader

_LOGGER = logging.getLogger(__name__)  # Logger for this module

DESCRIPTION_FILENAME = "variable_descriptions.txt"

# Metadata that will be ignore when checking variable equality and when adding variable
# to an OpenMDAO component
METADATA_TO_IGNORE = [
    "is_input",
    "tags",
    "size",
    "src_indices",
    "src_slice",
    "flat_src_indices",
    "distributed",
    "res_units",  # deprecated in IndepVarComp.add_output() since OpenMDAO 3.2
    "lower",  # deprecated in IndepVarComp.add_output() since OpenMDAO 3.2
    "upper",  # deprecated in IndepVarComp.add_output() since OpenMDAO 3.2
    "ref",  # deprecated in IndepVarComp.add_output() since OpenMDAO 3.2
    "ref0",  # deprecated in IndepVarComp.add_output() since OpenMDAO 3.2
    "res_ref",  # deprecated in IndepVarComp.add_output() since OpenMDAO 3.2
    "ref",  # deprecated in IndepVarComp.add_output() since OpenMDAO 3.2
    "global_shape",
    "global_size",
    "discrete",  # currently inconsistent in openMDAO 3.4
    "prom_name",
    "desc",
]


class Variable(Hashable):
    """
    A class for storing data of OpenMDAO variables.

    Instantiation is expected to be done through keyword arguments only.

    Beside the mandatory parameter 'name, kwargs is expected to have keys
    'value', 'units' and 'desc', that are accessible respectively through
    properties :meth:`name`, :meth:`value`, :meth:`units` and :meth:`description`.

    Other keys are possible. They match the definition of OpenMDAO's method
    :meth:`Component.add_output` described
    `here <http://openmdao.org/twodocs/versions/latest/_srcdocs/packages/core/
    component.html#openmdao.core.component.Component.add_output>`_.

    These keys can be listed with class method :meth:`get_openmdao_keys`.
    **Any other key in kwargs will be silently ignored.**

    Special behaviour: :meth:`description` will return the content of kwargs['desc']
    unless these 2 conditions are met:
     - kwargs['desc'] is None or 'desc' key is missing
     - a description exists in FAST-OAD internal data for the variable name
    Then, the internal description will be returned by :meth:`description`

    :param kwargs: the attributes of the variable, as keyword arguments
    """

    # Will store content of description files
    _variable_descriptions = {}

    # The list of modules of path where description files have been read
    _loaded_descriptions = set()

    # Default metadata
    _base_metadata = {}

    def __init__(self, name, **kwargs):
        super().__init__()

        self.name = name
        """ Name of the variable """

        self.metadata: Dict = {}
        """ Dictionary for metadata of the variable """

        # Initialize class attributes once at first instantiation -------------
        if not self._base_metadata:
            # Get variable base metadata from an ExplicitComponent
            comp = om.ExplicitComponent()
            # get attributes
            metadata = comp.add_output(name="a")

            self.__class__._base_metadata = metadata
            self.__class__._base_metadata["val"] = 1.0
            self.__class__._base_metadata["tags"] = set()
            self.__class__._base_metadata["shape"] = None
        # Done with class attributes ------------------------------------------

        # Feed self.metadata with kwargs, but remove first attributes with "Unavailable" as
        # value, which is a value that can be provided by OpenMDAO.
        self.metadata = self.__class__._base_metadata.copy()
        self.metadata.update(
            {
                key: value
                for key, value in kwargs.items()
                # The isinstance check is needed if value is a numpy array. In this case, a
                # FutureWarning is issued because it is compared to a scalar.
                if not isinstance(value, str) or value != "Unavailable"
            }
        )

        if "value" in self.metadata:
            self.metadata["val"] = self.metadata.pop("value")
        if "description" in self.metadata:
            self.metadata["desc"] = self.metadata.pop("description")

        self._set_default_shape()

        # If no description, use the one from self._variable_descriptions, if available
        if not self.description and self.name in self._variable_descriptions:
            self.description = self._variable_descriptions[self.name]

    @classmethod
    def read_variable_descriptions(cls, file_parent: str, update_existing: bool = True):
        """
        Reads variable descriptions in indicated folder or package, if it contains some.

        The file variable_descriptions.txt is looked for. Nothing is done if it is not
        found (no error raised also).

        Each line of the file should be formatted like::

            my:variable||The description of my:variable, as long as needed, but on one line.

        :param file_parent: the folder path or the package name that should contain the file
        :param update_existing: if True, previous descriptions will be updated.
                                if False, previous descriptions will be erased.
        """
        if not update_existing:
            cls._variable_descriptions = {}
            cls._loaded_descriptions = set()

        variable_descriptions = None
        description_file = None

        if file_parent:
            if pth.isdir(file_parent):
                file_path = pth.join(file_parent, DESCRIPTION_FILENAME)
                if pth.isfile(file_path):
                    description_file = open(file_path)
            else:
                # Then it is a module name
                if DESCRIPTION_FILENAME in PackageReader(file_parent).contents:
                    description_file = open_text(file_parent, DESCRIPTION_FILENAME)

        if description_file is not None:
            try:
                variable_descriptions = np.genfromtxt(
                    description_file, delimiter="||", dtype=str, autostrip=True
                )
            except Exception as exc:
                # Reading the file is not mandatory, so let's just log the error.
                _LOGGER.error(
                    "Could not read file %s in %s. Error log is:\n%s",
                    DESCRIPTION_FILENAME,
                    file_parent,
                    exc,
                )
            description_file.close()

        if variable_descriptions is not None:
            if np.shape(variable_descriptions) == (2,):
                # If the file contains only one line, np.genfromtxt() will return a (2,)-shaped
                # array. We need a reshape for dict.update() to work correctly.
                variable_descriptions = np.reshape(variable_descriptions, (1, 2))

            cls._loaded_descriptions.add(file_parent)
            cls.update_variable_descriptions(variable_descriptions)
            _LOGGER.info("Loaded variable descriptions in %s", file_parent)

    @classmethod
    def update_variable_descriptions(
        cls, variable_descriptions: Union[Mapping[str, str], Iterable[Tuple[str, str]]]
    ):
        """
        Updates description of variables.

        :param variable_descriptions: dict-like object with variable names as keys and descriptions
                                      as values
        """
        cls._variable_descriptions.update(variable_descriptions)

    @classmethod
    def get_openmdao_keys(cls):
        """

        :return: the keys that are used in OpenMDAO variables
        """
        # As _base_metadata is initialized at first instantiation, we create an instance to
        # ensure it has been done
        cls("dummy")

        return cls._base_metadata.keys()

    @property
    def value(self):
        """value of the variable"""
        return self.metadata.get("val")

    @value.setter
    def value(self, value):
        self.metadata["val"] = value
        self._set_default_shape()

    @property
    def val(self):
        """value of the variable (alias of property "value")"""
        return self.value

    @val.setter
    def val(self, value):
        self.value = value

    @property
    def units(self):
        """units associated to value (or None if not found)"""
        return self.metadata.get("units")

    @units.setter
    def units(self, value):
        self.metadata["units"] = value

    @property
    def description(self):
        """description of the variable (or None if not found)"""
        return self.metadata.get("desc")

    @description.setter
    def description(self, value):
        self.metadata["desc"] = value

    @property
    def desc(self):
        """description of the variable (or None if not found) (alias of property "description")"""
        return self.description

    @desc.setter
    def desc(self, value):
        self.description = value

    @property
    def is_input(self):
        """I/O status of the variable.

        - True if variable is a problem input
        - False if it is an output
        - None if information not found
        """
        return self.metadata.get("is_input")

    @is_input.setter
    def is_input(self, value):
        self.metadata["is_input"] = value

    def get_openmdao_kwargs(self, keys: Iterable = None) -> dict:
        """
        Provides a dict usable as keyword args by OpenMDAO add_input()/add_output().

        The dict keys will be the ones provided, or a default set if no keys are provided.

        :param keys:
        :return: the kwargs dict
        """
        if not keys:
            keys = {
                "name",
                "val",
                "units",
                "desc",
            }
            if self.metadata["shape_by_conn"] or self.metadata["copy_shape"]:
                keys.add("shape_by_conn")
                keys.add("copy_shape")
            else:
                keys.add("shape")

        kwargs = {key: self.metadata[key] for key in self.get_openmdao_keys() if key in keys}
        kwargs["name"] = self.name
        return kwargs

    def _set_default_shape(self):
        """Automatically sets shape if not set"""
        if self.metadata["shape"] is None:
            shape = np.shape(self.value)
            if not shape:
                shape = (1,)
            self.metadata["shape"] = shape

    def __eq__(self, other):
        # same arrays with nan are declared non equals, so we need a workaround
        my_metadata = dict(self.metadata)
        other_metadata = dict(other.metadata)
        my_value = np.asarray(my_metadata.pop("val"))
        other_value = np.asarray(other_metadata.pop("val"))

        # Let's also ignore unimportant keys
        for key in METADATA_TO_IGNORE:
            if key in my_metadata:
                del my_metadata[key]
            if key in other_metadata:
                del other_metadata[key]

        return (
            isinstance(other, Variable)
            and self.name == other.name
            and np.all(np.isclose(my_value, other_value, equal_nan=True))
            and my_metadata == other_metadata
        )

    def __repr__(self):
        return "Variable(name=%s, metadata=%s)" % (self.name, self.metadata)

    def __hash__(self) -> int:
        return hash("var=" + self.name)  # Name is normally unique



import numpy as np
import scipy.constants as sc


def compute_fuel_mission(owe, payload, mission_range, tsfc, l_d, cruise_speed):
    """
    Computes the fuel consumed during the mission based on the Breguet's range equation

    :param owe: the structural mass, in kg
    :param payload: the payload mass, in kg
    :param mission_range: the mission range, in m
    :param tsfc: the thrust specific fuel consumption, in kg/N/s
    :param l_d: the lift-to-drag ratio in cruise conditions, no unit
    :param cruise_speed: Cruise speed, in m/s

    :return mission_fuel: the fuel consumed during the designated mission, in kg
    """

    # To simplify the computation, we will first start by computing the range parameter,
    # which correspond to the term inside the exponential in the original formula
    range_parameter = (mission_range * tsfc * sc.g) / (cruise_speed * l_d)

    # Let's now computed the fuel using Breguet's range equation rearranged
    mission_fuel = (owe + payload) * (np.exp(range_parameter) - 1.0)

    return mission_fuel



"""
Defines the variable viewer for postprocessing
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from math import isnan
from typing import Dict

import ipysheet as sh
import ipywidgets as widgets
import numpy as np
import pandas as pd
from IPython.display import clear_output, display

from fastoad.io import DataFile
from fastoad.io.configuration.configuration import (
    FASTOADProblemConfigurator,
    KEY_CONSTRAINTS,
    KEY_DESIGN_VARIABLES,
    KEY_OBJECTIVE,
)
from fastoad.openmdao.variables import Variable, VariableList
from .exceptions import FastMissingFile

pd.set_option("display.max_rows", None)


class OptimizationViewer:
    """
    A class for interacting with FAST-OAD Problem optimization information.
    """

    # When getting a dataframe from a VariableList, the dictionary keys tell what columns
    #  are kept and values tell what name will be displayed.
    _DEFAULT_COLUMN_RENAMING = {
        "type": "Type",
        "name": "Name",
        "initial_value": "Initial Value",
        "lower": "Lower",
        "val": "Value",
        "upper": "Upper",
        "units": "Unit",
        "desc": "Description",
    }

    def __init__(self):

        #: Instance of the FAST-OAD problem configuration
        self.problem_configuration: FASTOADProblemConfigurator = None

        #: The dataframe which is the mirror of self.file
        self.dataframe = pd.DataFrame()

        # The sheet which is the mirror of the design var sheet
        self._design_var_sheet = None

        # The sheet which is the mirror of the constraint sheet
        self._constraint_sheet = None

        # The sheet which is the mirror of the objective sheet
        self._objective_sheet = None

        # The ui containing save and load buttons
        self._save_load_buttons = None

        # True if in the absence of an output file
        self._MISSING_OUTPUT_FILE = None

    def load(self, problem_configuration: FASTOADProblemConfigurator):
        """
        Loads the FAST-OAD problem and stores its data.

        :param problem_configuration: the FASTOADProblem instance.
        """

        self.problem_configuration = problem_configuration

        if pth.isfile(self.problem_configuration.input_file_path):
            input_variables = DataFile(self.problem_configuration.input_file_path)
        else:
            # TODO: generate the input file by default ?
            raise FastMissingFile("Please generate input file before using the optimization viewer")

        if pth.isfile(self.problem_configuration.output_file_path):
            self._MISSING_OUTPUT_FILE = False
            output_variables = DataFile(self.problem_configuration.output_file_path)
        else:
            self._MISSING_OUTPUT_FILE = True
            problem = self.problem_configuration.get_problem()
            problem.setup()
            output_variables = VariableList.from_problem(problem)

        optimization_variables = VariableList()
        opt_def = problem_configuration.get_optimization_definition()
        # Design Variables
        if KEY_DESIGN_VARIABLES in opt_def:
            for name, design_var in opt_def[KEY_DESIGN_VARIABLES].items():
                metadata = {
                    "type": "design_var",
                    "initial_value": input_variables[name].value,
                    "lower": design_var.get("lower"),
                    "value": output_variables[name].value,
                    "upper": design_var.get("upper"),
                    "units": input_variables[name].units,
                    "desc": input_variables[name].description,
                }
                optimization_variables[name] = metadata

        # Constraints
        if KEY_CONSTRAINTS in opt_def:
            for name, constr in opt_def[KEY_CONSTRAINTS].items():
                metadata = {
                    "type": "constraint",
                    "initial_value": None,
                    "lower": constr.get("lower"),
                    "value": output_variables[name].value,
                    "upper": constr.get("upper"),
                    "units": output_variables[name].units,
                    "desc": output_variables[name].description,
                }
                optimization_variables[name] = metadata

        # Objectives
        if KEY_OBJECTIVE in opt_def:
            for name in opt_def[KEY_OBJECTIVE]:
                metadata = {
                    "type": "objective",
                    "initial_value": None,
                    "lower": None,
                    "value": output_variables[name].value,
                    "upper": None,
                    "units": output_variables[name].units,
                    "desc": output_variables[name].description,
                }
                optimization_variables[name] = metadata

        self.load_variables(optimization_variables)

    def save(self):
        """
        Save the optimization to the files.
        Possible files modified are:
            - the .yml configuration file
            - the input file (initial values)
            - the output file (values)
        """
        conf = self.problem_configuration
        input_variables = DataFile(self.problem_configuration.input_file_path, None)
        opt_def = conf.get_optimization_definition()

        variables = self.get_variables()
        for variable in variables:
            name = variable.name
            if name in input_variables.names():
                input_variables[name].value = variable.metadata["initial_value"]
            self._update_optim_variable(variable, opt_def)

        # Saving modifications
        # Initial values
        input_variables.save()

        # Optimization definition
        conf.set_optimization_definition(opt_def)
        conf.save()

    @staticmethod
    def _update_optim_variable(variable: Variable, optim_definition: Dict):
        """
        Updates optim_definition with metadata of provided variable.

        :param variable:
        :param optim_definition:
        """
        name = variable.name
        meta = variable.metadata

        if meta["type"] == "design_var":
            # TODO: later it will be possible to add/remove design variables in the ui
            section_name = KEY_DESIGN_VARIABLES
        elif meta["type"] == "constraint":
            # TODO: later it will be possible to add/remove constraints in the ui
            section_name = KEY_CONSTRAINTS
        else:
            return

        if section_name not in optim_definition:
            optim_definition[section_name] = {}
        if name not in optim_definition[section_name]:
            optim_definition[section_name][name] = {}
        for bound in ["lower", "upper"]:
            if meta[bound] is not None and not isnan(meta[bound]):
                optim_definition[section_name][name].update({bound: meta[bound]})
            else:
                optim_definition[section_name][name].pop(bound, None)

    def display(self):
        """
        Displays the datasheet.
        load() must be ran before.

        :return: display of the user interface:
        """
        self._create_save_load_buttons()
        return self._render_ui()

    def load_variables(self, variables: VariableList, attribute_to_column: Dict[str, str] = None):
        """
        Loads provided variable list and replace current data set.

        :param variables: the variables to load
        :param attribute_to_column: dictionary keys tell what variable attributes are
               kept and the values tell what name will be displayed. If not provided,
               default translation will apply.
        """

        if not attribute_to_column:
            attribute_to_column = self._DEFAULT_COLUMN_RENAMING

        self.dataframe = (
            variables.to_dataframe()
            .rename(columns=attribute_to_column)[attribute_to_column.values()]
            .reset_index(drop=True)
        )

    def get_variables(self, column_to_attribute: Dict[str, str] = None) -> VariableList:
        """

        :param column_to_attribute: dictionary keys tell what columns are kept and the values
                                    tell whatvariable attribute it corresponds to. If not
                                    provided, default translation will apply.
        :return: a variable list from current data set
        """
        if not column_to_attribute:
            column_to_attribute = {
                value: key for key, value in self._DEFAULT_COLUMN_RENAMING.items()
            }

        return VariableList.from_dataframe(
            self.dataframe[column_to_attribute.keys()].rename(columns=column_to_attribute)
        )

    # pylint: disable=invalid-name # df is a common naming for dataframes
    def _df_to_sheet(self, df: pd.DataFrame) -> sh.Sheet:
        """
        Transforms a pandas DataFrame into a ipysheet Sheet.
        The cells are set to read only except for the values.

        :param df: the pandas DataFrame to be converted
        :return: the equivalent ipysheet Sheet
        """
        if not df.empty:
            # Adapted from_dataframe() method of ipysheet
            columns = df.columns.tolist()
            rows = df.index.tolist()
            cells = []

            read_only_cells = ["Name", "Unit", "Description", "Value"]

            style = self._cell_styling(df)
            row_idx = 0
            for r in rows:
                col_idx = 0
                for c in columns:
                    value = df.loc[r, c]
                    if c in read_only_cells:
                        read_only = True
                        numeric_format = None
                    else:
                        read_only = False
                        # TODO: make the number of decimals depend on the module ?
                        # or chosen in the ui by the user
                        numeric_format = "0.000"

                    # If no output file is provided make it clearer for the user
                    if c == "Value" and self._MISSING_OUTPUT_FILE:
                        value = "-"

                    cells.append(
                        sh.Cell(
                            value=value,
                            row_start=row_idx,
                            row_end=row_idx,
                            column_start=col_idx,
                            column_end=col_idx,
                            numeric_format=numeric_format,
                            read_only=read_only,
                            style=style[(r, c)],
                        )
                    )
                    col_idx += 1
                row_idx += 1
            sheet = sh.Sheet(
                rows=len(rows),
                columns=len(columns),
                cells=cells,
                row_headers=[str(header) for header in rows],
                column_headers=[str(header) for header in columns],
            )

        else:
            sheet = sh.sheet(rows=0, columns=0)

        return sheet

    @staticmethod
    def _sheet_to_df(sheet: sh.Sheet) -> pd.DataFrame:
        """
        Transforms a ipysheet Sheet into a pandas DataFrame.

        :param sheet: the ipysheet Sheet to be converted
        :return: the equivalent pandas DataFrame
        """
        df = sh.to_dataframe(sheet)
        return df

    # pylint: disable=unused-argument  # args has to be there for observe() to work
    def _update_df(self, change=None):
        """
        Updates the stored DataFrame with respect to the actual values of the Sheet.
        Then updates the file with respect to the stored DataFrame.
        """
        frames = [
            self._sheet_to_df(self._design_var_sheet),
            self._sheet_to_df(self._constraint_sheet),
            self._sheet_to_df(self._objective_sheet),
        ]

        df = pd.concat(frames, sort=True)
        columns = {}
        columns.update(self._DEFAULT_COLUMN_RENAMING)
        columns.pop("type")

        column_to_attribute = {value: key for key, value in columns.items()}

        variables = VariableList.from_dataframe(
            df[column_to_attribute.keys()].rename(columns=column_to_attribute)
        )

        attribute_to_column = columns
        df = (
            variables.to_dataframe()
            .rename(columns=attribute_to_column)[attribute_to_column.values()]
            .reset_index(drop=True)
        )
        rows = df.index.tolist()
        columns = df.columns.tolist()

        for r in rows:
            for c in columns:
                self.dataframe.loc[int(r), c] = df.loc[r, c]

    def _create_save_load_buttons(self):
        """
        The save button saves the present state of the dataframe to the xml.
        The load button loads the xml and replaces actual the dataframe.
        """

        save_button = widgets.Button(
            description="Save",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Save to the file",
            icon="save",
        )

        def on_save_button_clicked(b):
            self.save()

        save_button.on_click(on_save_button_clicked)

        load_button = widgets.Button(
            description="Load",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Load the file",
            icon="upload",
        )

        def on_load_button_clicked(b):
            self.load(self.problem_configuration)
            self._render_ui()

        load_button.on_click(on_load_button_clicked)

        items_box = widgets.HBox([load_button, save_button])

        self._save_load_buttons = items_box

    def _update_sheet(self):
        """
        Updates the sheet after filtering the dataframe with respect to
        the actual values of the variable dropdown menus.
        """
        design_var_df = self.dataframe[self.dataframe["Type"] == "design_var"]
        design_var_df = design_var_df.drop(columns=["Type"])
        self._design_var_sheet = self._df_to_sheet(design_var_df)
        constraint_df = self.dataframe[self.dataframe["Type"] == "constraint"]
        constraint_df = constraint_df.drop(columns=["Type", "Initial Value"])
        self._constraint_sheet = self._df_to_sheet(constraint_df)
        objective_df = self.dataframe[self.dataframe["Type"] == "objective"]
        objective_df = objective_df.drop(columns=["Type", "Initial Value", "Lower", "Upper"])
        self._objective_sheet = self._df_to_sheet(objective_df)

        for cell in self._design_var_sheet.cells:
            cell.observe(self._update_df, "value")
            cell.observe(self._update_style, "value")

        for cell in self._constraint_sheet.cells:
            cell.observe(self._update_df, "value")
            cell.observe(self._update_style, "value")

        for cell in self._objective_sheet.cells:
            cell.observe(self._update_df, "value")
            cell.observe(self._update_style, "value")

    # pylint: disable=unused-argument  # args has to be there for observe() to work
    def _render_ui(self, change=None) -> display:
        """
        Renders the dropdown menus for the variable selector and the corresponding
        ipysheet Sheet containing the variable infos.

        :return: the display object
        """
        clear_output(wait=True)
        self._update_sheet()
        ui = widgets.VBox(
            [
                self._save_load_buttons,
                self._design_var_ui(),
                self._constraint_ui(),
                self._objective_ui(),
            ]
        )
        return display(ui)

    def _design_var_ui(self):
        return widgets.VBox([widgets.Label(value="Design Variables"), self._design_var_sheet])

    def _constraint_ui(self):
        return widgets.VBox([widgets.Label(value="Constraints"), self._constraint_sheet])

    def _objective_ui(self):
        return widgets.VBox([widgets.Label(value="Objectives"), self._objective_sheet])

    @staticmethod
    def _cell_styling(df) -> Dict:
        """
        Returns bound activities in the form of cell style dictionary.

        :return: dict containing the style
        """

        def highlight_active_bounds(df, threshold=1e-6):
            rows = df.index.tolist()
            columns = df.columns.tolist()
            style = {}
            for r in rows:
                s = df.loc[r]
                is_active = pd.Series(data=False, index=s.index)
                is_violated = pd.Series(data=False, index=s.index)
                if "Lower" in s:
                    # Constraints might only have a upper bound
                    if s.loc["Lower"] is not None:
                        if np.all(s.loc["Lower"] + threshold >= s.loc["Value"]) & np.all(
                            s.loc["Value"] >= s.loc["Lower"] - threshold
                        ):
                            is_active["Lower"] = True
                            is_active["Value"] = True
                        elif np.all(s.loc["Value"] < s.loc["Lower"] - threshold):
                            is_violated["Lower"] = True
                            is_violated["Value"] = True
                        else:
                            pass

                if "Upper" in s:
                    # Constraints might only have a lower bound
                    if s.loc["Upper"] is not None:
                        if np.all(s.loc["Upper"] + threshold >= s.loc["Value"]) & np.all(
                            s.loc["Value"] >= s.loc["Upper"] - threshold
                        ):
                            is_active["Upper"] = True
                            is_active["Value"] = True
                        elif np.all(s.loc["Value"] > s.loc["Upper"] + threshold):
                            is_violated["Upper"] = True
                            is_violated["Value"] = True
                        else:
                            pass

                yellow = ["yellow" if v else None for v in is_active]
                red = ["red" if v else None for v in is_violated]
                column_styles = [
                    {"backgroundColor": y_style or r_style} for y_style, r_style in zip(yellow, red)
                ]

                for column, column_style in zip(columns, column_styles):
                    style[(r, column)] = column_style

            return style

        style = highlight_active_bounds(df, threshold=0.1)
        # style.update(another_styling_method())

        return style

    def _update_style(self, change=None):
        """
        Updates the style of the sheet cells with respect to bound activities
        of the actual self.dataframe.
        """
        # Design variables
        design_var_df = self.dataframe[self.dataframe["Type"] == "design_var"]
        design_var_df = design_var_df.drop(columns=["Type"])
        design_var_df = design_var_df.reset_index(drop=True)
        style = self._cell_styling(design_var_df)

        for cell in self._design_var_sheet.cells:
            i, j = cell.row_start, cell.column_start
            r, c = (design_var_df.index.to_list()[i], design_var_df.columns.to_list()[j])
            cell.style = style[(r, c)]

        #  Constraints
        constraint_df = self.dataframe[self.dataframe["Type"] == "constraint"]
        constraint_df = constraint_df.drop(columns=["Type", "Initial Value"])
        constraint_df = constraint_df.reset_index(drop=True)
        style = self._cell_styling(constraint_df)

        for cell in self._constraint_sheet.cells:
            i, j = cell.row_start, cell.column_start
            r, c = (constraint_df.index.to_list()[i], constraint_df.columns.to_list()[j])
            cell.style = style[(r, c)]

        # Objectives
        objective_df = self.dataframe[self.dataframe["Type"] == "objective"]
        objective_df = objective_df.drop(columns=["Type", "Initial Value", "Lower", "Upper"])
        objective_df = objective_df.reset_index(drop=True)
        style = self._cell_styling(objective_df)

        for cell in self._objective_sheet.cells:
            i, j = cell.row_start, cell.column_start
            r, c = (objective_df.index.to_list()[i], objective_df.columns.to_list()[j])
            cell.style = style[(r, c)]



"""Aerodynamic polar data."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numpy import ndarray
from scipy.interpolate import interp1d
from scipy.optimize import fmin


class Polar:
    def __init__(self, cl: ndarray, cd: ndarray):

        """
        Class for managing aerodynamic polar data.

        Links drag coefficient (CD) to lift coefficient (CL).
        It is defined by two vectors with CL and CD values.

        Once defined, for any CL value, CD can be obtained using :meth:`cd`.

        :param cl: a N-elements array with CL values
        :param cd: a N-elements array with CD values that match CL
        """
        self._definition_CL = cl
        self._cd = interp1d(cl, cd, kind="quadratic", fill_value="extrapolate")

        def _negated_lift_drag_ratio(lift_coeff):
            """Returns -CL/CD."""
            return -lift_coeff / self.cd(lift_coeff)

        self._optimal_CL = fmin(_negated_lift_drag_ratio, cl[0], disp=0)

    @property
    def definition_cl(self):
        """The vector that has been used for defining lift coefficient."""
        return self._definition_CL

    @property
    def optimal_cl(self):
        """The CL value that provides larger lift/drag ratio."""
        return self._optimal_CL

    def cd(self, cl=None):
        """
        Computes drag coefficient (CD) by interpolation in definition data.

        :param cl: lift coefficient (CL) values. If not provided, the CL definition vector will be
                   used (i.e. CD definition vector will be returned)
        :return: CD values for each provide CL values
        """
        if cl is None:
            return self._cd(self._definition_CL)
        return self._cd(cl)



import numpy as np

import openmdao.api as om


class ComputeInducedDragCoefficient(om.ExplicitComponent):
    """
    Computes the induced drag coefficient based on the aspect ratio
    """

    def setup(self):

        # Defining the input(s)

        self.add_input(name="aspect_ratio", val=np.nan)

        # Defining the output(s)

        self.add_output(name="induced_drag_coefficient")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        # Assigning the input to local variable for clarity
        aspect_ratio = inputs["aspect_ratio"]

        # Computation of the Oswald efficiency factor
        e = 1.78 * (1.0 - 0.045 * aspect_ratio ** 0.68) - 0.64

        # Computation of the lift induced drag coefficient
        k = 1.0 / (np.pi * aspect_ratio * e)

        outputs["induced_drag_coefficient"] = k



""" Exceptions for io.xml module """

#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.exceptions import FastError


class FastXPathEvalError(FastError):
    """
    Raised when some xpath could not be resolved
    """


class FastXpathTranslatorInconsistentLists(FastError):
    """
    Raised when list of variable names and list of XPaths have not the same length
    """


class FastXpathTranslatorDuplicates(FastError):
    """
    Raised when list of variable names or list of XPaths have duplicate entries
    """


class FastXpathTranslatorVariableError(FastError):
    """
    Raised when a variable does not match any xpath in the translator file.
    """

    def __init__(self, variable):
        super().__init__("Unknown variable %s" % variable)
        self.variable = variable


class FastXpathTranslatorXPathError(FastError):
    """
    Raised when a xpath does not match any variable in the translator file.
    """

    def __init__(self, xpath):
        super().__init__("Unknown xpath %s" % xpath)
        self.xpath = xpath


class FastXmlFormatterDuplicateVariableError(FastError):
    """
    Raised a variable is defined more than once in a XML file
    """



import openmdao.api as om

from ...OpenMDAO.geometry.sub_components.compute_wing_area import ComputeWingArea

import fastoad.api as oad


@oad.RegisterOpenMDAOSystem("tutorial.fast_oad.geometry")
class ComputeGeometry(om.Group):
    def setup(self):

        self.add_subsystem(name="compute_wing_area", subsys=ComputeWingArea(), promotes=["*"])



"""
Module for string-related operations
"""

#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import io
import re

import numpy as np

from fastoad.exceptions import FastError


def get_float_list_from_string(text: str):
    """
    Parses the provided string and returns a list of floats if possible.

    If provided text is not numeric, None is returned.

    New line characters are simply ignored.

    As an example, following patterns will result as list [1., 2., 3.]

    .. code-block::

        '[ 1, 2., 3]'
        ' 1, 2., 3'
        ' 1 2  3'
        ' 1 2  3 dummy 4'
    """

    text_value = text.strip()
    if not text_value:
        return None

    # If it begins by '[', an array is expected, potentially multidimensional
    if text_value.startswith("["):
        # The string is first transformed in a way that can be parsed by genfromtxt
        text_value = re.sub(r"\r?\n|\r", "", text_value)  # first remove all new lines
        text_value = re.sub(r"\]\s*,\s*\[", "\n", text_value)
        text_value = text_value.strip("[]")
        text_io = io.StringIO(text_value)
        try:
            return np.genfromtxt(text_io, delimiter=",").tolist()
        except ValueError as exc:
            raise FastCouldNotParseStringToArrayError(text.strip(), exc)

    # Deals with multiple values in same element. numpy.fromstring can parse a string,
    # but we have to test with either ' ' or ',' as separator. The longest result should be
    # the good one.
    try:
        value1 = np.fromstring(text_value, dtype=float, sep=" ").tolist()
    except ValueError:
        pass

    try:
        value2 = np.fromstring(text_value, dtype=float, sep=",").tolist()
    except ValueError:
        pass

    if not value1 and not value2:
        return None

    return value1 if len(value1) > len(value2) else value2


class FastCouldNotParseStringToArrayError(FastError):
    """Raised when a conversion from string to array failed."""

    def __init__(self, parsed_text, original_exception):
        super().__init__()
        self.text = parsed_text
        self.original_exception = original_exception

    def __str__(self):
        msg = 'Could not parse "%s"' % self.text
        if self.original_exception:
            msg += ': got Error "%s"' % self.original_exception
        return msg
